# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import logging
import inspect
import struct

from os_ken.lib import addrconv
from os_ken.lib import ip
from os_ken.lib.packet import ipv6


LOG = logging.getLogger(__name__)


class Test_ipv6(unittest.TestCase):

    def setUp(self):
        self.version = 6
        self.traffic_class = 0
        self.flow_label = 0
        self.payload_length = 817
        self.nxt = 6
        self.hop_limit = 128
        self.src = '2002:4637:d5d3::4637:d5d3'
        self.dst = '2001:4860:0:2001::68'
        self.ext_hdrs = []
        self.ip = ipv6.ipv6(
            self.version, self.traffic_class, self.flow_label,
            self.payload_length, self.nxt, self.hop_limit, self.src,
            self.dst, self.ext_hdrs)

        self.v_tc_flow = (
            self.version << 28 | self.traffic_class << 20 |
            self.flow_label << 12)
        self.buf = struct.pack(
            ipv6.ipv6._PACK_STR, self.v_tc_flow,
            self.payload_length, self.nxt, self.hop_limit,
            addrconv.ipv6.text_to_bin(self.src),
            addrconv.ipv6.text_to_bin(self.dst))

    def setUp_with_hop_opts(self):
        self.opt1_type = 5
        self.opt1_len = 2
        self.opt1_data = b'\x00\x00'
        self.opt2_type = 1
        self.opt2_len = 0
        self.opt2_data = None
        self.options = [
            ipv6.option(self.opt1_type, self.opt1_len, self.opt1_data),
            ipv6.option(self.opt2_type, self.opt2_len, self.opt2_data),
        ]
        self.hop_opts_nxt = 6
        self.hop_opts_size = 0
        self.hop_opts = ipv6.hop_opts(
            self.hop_opts_nxt, self.hop_opts_size, self.options)
        self.ext_hdrs = [self.hop_opts]
        self.payload_length += len(self.hop_opts)
        self.nxt = ipv6.hop_opts.TYPE
        self.ip = ipv6.ipv6(
            self.version, self.traffic_class, self.flow_label,
            self.payload_length, self.nxt, self.hop_limit, self.src,
            self.dst, self.ext_hdrs)
        self.buf = struct.pack(
            ipv6.ipv6._PACK_STR, self.v_tc_flow,
            self.payload_length, self.nxt, self.hop_limit,
            addrconv.ipv6.text_to_bin(self.src),
            addrconv.ipv6.text_to_bin(self.dst))
        self.buf += self.hop_opts.serialize()

    def setUp_with_dst_opts(self):
        self.opt1_type = 5
        self.opt1_len = 2
        self.opt1_data = b'\x00\x00'
        self.opt2_type = 1
        self.opt2_len = 0
        self.opt2_data = None
        self.options = [
            ipv6.option(self.opt1_type, self.opt1_len, self.opt1_data),
            ipv6.option(self.opt2_type, self.opt2_len, self.opt2_data),
        ]
        self.dst_opts_nxt = 6
        self.dst_opts_size = 0
        self.dst_opts = ipv6.dst_opts(
            self.dst_opts_nxt, self.dst_opts_size, self.options)
        self.ext_hdrs = [self.dst_opts]
        self.payload_length += len(self.dst_opts)
        self.nxt = ipv6.dst_opts.TYPE
        self.ip = ipv6.ipv6(
            self.version, self.traffic_class, self.flow_label,
            self.payload_length, self.nxt, self.hop_limit, self.src,
            self.dst, self.ext_hdrs)
        self.buf = struct.pack(
            ipv6.ipv6._PACK_STR, self.v_tc_flow,
            self.payload_length, self.nxt, self.hop_limit,
            addrconv.ipv6.text_to_bin(self.src),
            addrconv.ipv6.text_to_bin(self.dst))
        self.buf += self.dst_opts.serialize()

    def setUp_with_routing_type3(self):
        self.routing_nxt = 6
        self.routing_size = 6
        self.routing_type = 3
        self.routing_seg = 2
        self.routing_cmpi = 0
        self.routing_cmpe = 0
        self.routing_adrs = ["2001:db8:dead::1", "2001:db8:dead::2",
                             "2001:db8:dead::3"]
        self.routing = ipv6.routing_type3(
            self.routing_nxt, self.routing_size,
            self.routing_type, self.routing_seg,
            self.routing_cmpi, self.routing_cmpe,
            self.routing_adrs)
        self.ext_hdrs = [self.routing]
        self.payload_length += len(self.routing)
        self.nxt = ipv6.routing.TYPE
        self.ip = ipv6.ipv6(
            self.version, self.traffic_class, self.flow_label,
            self.payload_length, self.nxt, self.hop_limit, self.src,
            self.dst, self.ext_hdrs)
        self.buf = struct.pack(
            ipv6.ipv6._PACK_STR, self.v_tc_flow,
            self.payload_length, self.nxt, self.hop_limit,
            addrconv.ipv6.text_to_bin(self.src),
            addrconv.ipv6.text_to_bin(self.dst))
        self.buf += self.routing.serialize()

    def setUp_with_fragment(self):
        self.fragment_nxt = 6
        self.fragment_offset = 50
        self.fragment_more = 1
        self.fragment_id = 123
        self.fragment = ipv6.fragment(
            self.fragment_nxt, self.fragment_offset, self.fragment_more,
            self.fragment_id)
        self.ext_hdrs = [self.fragment]
        self.payload_length += len(self.fragment)
        self.nxt = ipv6.fragment.TYPE
        self.ip = ipv6.ipv6(
            self.version, self.traffic_class, self.flow_label,
            self.payload_length, self.nxt, self.hop_limit, self.src,
            self.dst, self.ext_hdrs)
        self.buf = struct.pack(
            ipv6.ipv6._PACK_STR, self.v_tc_flow,
            self.payload_length, self.nxt, self.hop_limit,
            addrconv.ipv6.text_to_bin(self.src),
            addrconv.ipv6.text_to_bin(self.dst))
        self.buf += self.fragment.serialize()

    def setUp_with_auth(self):
        self.auth_nxt = 6
        self.auth_size = 4
        self.auth_spi = 256
        self.auth_seq = 1
        self.auth_data = b'\xa0\xe7\xf8\xab\xf9\x69\x1a\x8b\xf3\x9f\x7c\xae'
        self.auth = ipv6.auth(
            self.auth_nxt, self.auth_size, self.auth_spi, self.auth_seq,
            self.auth_data)
        self.ext_hdrs = [self.auth]
        self.payload_length += len(self.auth)
        self.nxt = ipv6.auth.TYPE
        self.ip = ipv6.ipv6(
            self.version, self.traffic_class, self.flow_label,
            self.payload_length, self.nxt, self.hop_limit, self.src,
            self.dst, self.ext_hdrs)
        self.buf = struct.pack(
            ipv6.ipv6._PACK_STR, self.v_tc_flow,
            self.payload_length, self.nxt, self.hop_limit,
            addrconv.ipv6.text_to_bin(self.src),
            addrconv.ipv6.text_to_bin(self.dst))
        self.buf += self.auth.serialize()

    def setUp_with_multi_headers(self):
        self.opt1_type = 5
        self.opt1_len = 2
        self.opt1_data = b'\x00\x00'
        self.opt2_type = 1
        self.opt2_len = 0
        self.opt2_data = None
        self.options = [
            ipv6.option(self.opt1_type, self.opt1_len, self.opt1_data),
            ipv6.option(self.opt2_type, self.opt2_len, self.opt2_data),
        ]
        self.hop_opts_nxt = ipv6.auth.TYPE
        self.hop_opts_size = 0
        self.hop_opts = ipv6.hop_opts(
            self.hop_opts_nxt, self.hop_opts_size, self.options)
        self.auth_nxt = 6
        self.auth_size = 4
        self.auth_spi = 256
        self.auth_seq = 1
        self.auth_data = b'\xa0\xe7\xf8\xab\xf9\x69\x1a\x8b\xf3\x9f\x7c\xae'
        self.auth = ipv6.auth(
            self.auth_nxt, self.auth_size, self.auth_spi, self.auth_seq,
            self.auth_data)
        self.ext_hdrs = [self.hop_opts, self.auth]
        self.payload_length += len(self.hop_opts) + len(self.auth)
        self.nxt = ipv6.hop_opts.TYPE
        self.ip = ipv6.ipv6(
            self.version, self.traffic_class, self.flow_label,
            self.payload_length, self.nxt, self.hop_limit, self.src,
            self.dst, self.ext_hdrs)
        self.buf = struct.pack(
            ipv6.ipv6._PACK_STR, self.v_tc_flow,
            self.payload_length, self.nxt, self.hop_limit,
            addrconv.ipv6.text_to_bin(self.src),
            addrconv.ipv6.text_to_bin(self.dst))
        self.buf += self.hop_opts.serialize()
        self.buf += self.auth.serialize()

    def tearDown(self):
        pass

    def test_init(self):
        self.assertEqual(self.version, self.ip.version)
        self.assertEqual(self.traffic_class, self.ip.traffic_class)
        self.assertEqual(self.flow_label, self.ip.flow_label)
        self.assertEqual(self.payload_length, self.ip.payload_length)
        self.assertEqual(self.nxt, self.ip.nxt)
        self.assertEqual(self.hop_limit, self.ip.hop_limit)
        self.assertEqual(self.src, self.ip.src)
        self.assertEqual(self.dst, self.ip.dst)
        self.assertEqual(str(self.ext_hdrs), str(self.ip.ext_hdrs))

    def test_init_with_hop_opts(self):
        self.setUp_with_hop_opts()
        self.test_init()

    def test_init_with_dst_opts(self):
        self.setUp_with_dst_opts()
        self.test_init()

    def test_init_with_routing_type3(self):
        self.setUp_with_routing_type3()
        self.test_init()

    def test_init_with_fragment(self):
        self.setUp_with_fragment()
        self.test_init()

    def test_init_with_auth(self):
        self.setUp_with_auth()
        self.test_init()

    def test_init_with_multi_headers(self):
        self.setUp_with_multi_headers()
        self.test_init()

    def test_parser(self):
        _res = self.ip.parser(bytes(self.buf))
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res

        self.assertEqual(self.version, res.version)
        self.assertEqual(self.traffic_class, res.traffic_class)
        self.assertEqual(self.flow_label, res.flow_label)
        self.assertEqual(self.payload_length, res.payload_length)
        self.assertEqual(self.nxt, res.nxt)
        self.assertEqual(self.hop_limit, res.hop_limit)
        self.assertEqual(self.src, res.src)
        self.assertEqual(self.dst, res.dst)
        self.assertEqual(str(self.ext_hdrs), str(res.ext_hdrs))

    def test_parser_with_hop_opts(self):
        self.setUp_with_hop_opts()
        self.test_parser()

    def test_parser_with_dst_opts(self):
        self.setUp_with_dst_opts()
        self.test_parser()

    def test_parser_with_routing_type3(self):
        self.setUp_with_routing_type3()
        self.test_parser()

    def test_parser_with_fragment(self):
        self.setUp_with_fragment()
        self.test_parser()

    def test_parser_with_auth(self):
        self.setUp_with_auth()
        self.test_parser()

    def test_parser_with_multi_headers(self):
        self.setUp_with_multi_headers()
        self.test_parser()

    def test_serialize(self):
        data = bytearray()
        prev = None
        buf = self.ip.serialize(data, prev)

        res = struct.unpack_from(ipv6.ipv6._PACK_STR, bytes(buf))

        self.assertEqual(self.v_tc_flow, res[0])
        self.assertEqual(self.payload_length, res[1])
        self.assertEqual(self.nxt, res[2])
        self.assertEqual(self.hop_limit, res[3])
        self.assertEqual(self.src, addrconv.ipv6.bin_to_text(res[4]))
        self.assertEqual(self.dst, addrconv.ipv6.bin_to_text(res[5]))

    def test_serialize_with_hop_opts(self):
        self.setUp_with_hop_opts()
        self.test_serialize()

        data = bytearray()
        prev = None
        buf = self.ip.serialize(data, prev)
        hop_opts = ipv6.hop_opts.parser(bytes(buf[ipv6.ipv6._MIN_LEN:]))
        self.assertEqual(repr(self.hop_opts), repr(hop_opts))

    def test_serialize_with_dst_opts(self):
        self.setUp_with_dst_opts()
        self.test_serialize()

        data = bytearray()
        prev = None
        buf = self.ip.serialize(data, prev)
        dst_opts = ipv6.dst_opts.parser(bytes(buf[ipv6.ipv6._MIN_LEN:]))
        self.assertEqual(repr(self.dst_opts), repr(dst_opts))

    def test_serialize_with_routing_type3(self):
        self.setUp_with_routing_type3()
        self.test_serialize()

        data = bytearray()
        prev = None
        buf = self.ip.serialize(data, prev)
        routing = ipv6.routing.parser(bytes(buf[ipv6.ipv6._MIN_LEN:]))
        self.assertEqual(repr(self.routing), repr(routing))

    def test_serialize_with_fragment(self):
        self.setUp_with_fragment()
        self.test_serialize()

        data = bytearray()
        prev = None
        buf = self.ip.serialize(data, prev)
        fragment = ipv6.fragment.parser(bytes(buf[ipv6.ipv6._MIN_LEN:]))
        self.assertEqual(repr(self.fragment), repr(fragment))

    def test_serialize_with_auth(self):
        self.setUp_with_auth()
        self.test_serialize()

        data = bytearray()
        prev = None
        buf = self.ip.serialize(data, prev)
        auth = ipv6.auth.parser(bytes(buf[ipv6.ipv6._MIN_LEN:]))
        self.assertEqual(repr(self.auth), repr(auth))

    def test_serialize_with_multi_headers(self):
        self.setUp_with_multi_headers()
        self.test_serialize()

        data = bytearray()
        prev = None
        buf = self.ip.serialize(data, prev)
        offset = ipv6.ipv6._MIN_LEN
        hop_opts = ipv6.hop_opts.parser(bytes(buf[offset:]))
        offset += len(hop_opts)
        auth = ipv6.auth.parser(bytes(buf[offset:]))
        self.assertEqual(repr(self.hop_opts), repr(hop_opts))
        self.assertEqual(repr(self.auth), repr(auth))

    def test_to_string(self):
        ipv6_values = {'version': self.version,
                       'traffic_class': self.traffic_class,
                       'flow_label': self.flow_label,
                       'payload_length': self.payload_length,
                       'nxt': self.nxt,
                       'hop_limit': self.hop_limit,
                       'src': repr(self.src),
                       'dst': repr(self.dst),
                       'ext_hdrs': self.ext_hdrs}
        _ipv6_str = ','.join(['%s=%s' % (k, ipv6_values[k])
                              for k, v in inspect.getmembers(self.ip)
                              if k in ipv6_values])
        ipv6_str = '%s(%s)' % (ipv6.ipv6.__name__, _ipv6_str)

        self.assertEqual(str(self.ip), ipv6_str)
        self.assertEqual(repr(self.ip), ipv6_str)

    def test_to_string_with_hop_opts(self):
        self.setUp_with_hop_opts()
        self.test_to_string()

    def test_to_string_with_dst_opts(self):
        self.setUp_with_dst_opts()
        self.test_to_string()

    def test_to_string_with_fragment(self):
        self.setUp_with_fragment()
        self.test_to_string()

    def test_to_string_with_auth(self):
        self.setUp_with_auth()
        self.test_to_string()

    def test_to_string_with_multi_headers(self):
        self.setUp_with_multi_headers()
        self.test_to_string()

    def test_len(self):
        self.assertEqual(len(self.ip), 40)

    def test_len_with_hop_opts(self):
        self.setUp_with_hop_opts()
        self.assertEqual(len(self.ip), 40 + len(self.hop_opts))

    def test_len_with_dst_opts(self):
        self.setUp_with_dst_opts()
        self.assertEqual(len(self.ip), 40 + len(self.dst_opts))

    def test_len_with_routing_type3(self):
        self.setUp_with_routing_type3()
        self.assertEqual(len(self.ip), 40 + len(self.routing))

    def test_len_with_fragment(self):
        self.setUp_with_fragment()
        self.assertEqual(len(self.ip), 40 + len(self.fragment))

    def test_len_with_auth(self):
        self.setUp_with_auth()
        self.assertEqual(len(self.ip), 40 + len(self.auth))

    def test_len_with_multi_headers(self):
        self.setUp_with_multi_headers()
        self.assertEqual(len(self.ip), 40 + len(self.hop_opts) + len(self.auth))

    def test_default_args(self):
        ip = ipv6.ipv6()
        buf = ip.serialize(bytearray(), None)
        res = struct.unpack(ipv6.ipv6._PACK_STR, bytes(buf))

        self.assertEqual(res[0], 6 << 28)
        self.assertEqual(res[1], 0)
        self.assertEqual(res[2], 6)
        self.assertEqual(res[3], 255)
        self.assertEqual(res[4], addrconv.ipv6.text_to_bin('10::10'))
        self.assertEqual(res[5], addrconv.ipv6.text_to_bin('20::20'))

        # with extension header
        ip = ipv6.ipv6(
            nxt=0, ext_hdrs=[
                ipv6.hop_opts(58, 0, [
                    ipv6.option(5, 2, b'\x00\x00'),
                    ipv6.option(1, 0, None)])])
        buf = ip.serialize(bytearray(), None)
        res = struct.unpack(ipv6.ipv6._PACK_STR + '8s', bytes(buf))

        self.assertEqual(res[0], 6 << 28)
        self.assertEqual(res[1], 8)
        self.assertEqual(res[2], 0)
        self.assertEqual(res[3], 255)
        self.assertEqual(res[4], addrconv.ipv6.text_to_bin('10::10'))
        self.assertEqual(res[5], addrconv.ipv6.text_to_bin('20::20'))
        self.assertEqual(res[6], b'\x3a\x00\x05\x02\x00\x00\x01\x00')

    def test_json(self):
        jsondict = self.ip.to_jsondict()
        ip = ipv6.ipv6.from_jsondict(jsondict['ipv6'])
        self.assertEqual(str(self.ip), str(ip))

    def test_json_with_hop_opts(self):
        self.setUp_with_hop_opts()
        self.test_json()

    def test_json_with_dst_opts(self):
        self.setUp_with_dst_opts()
        self.test_json()

    def test_json_with_routing_type3(self):
        self.setUp_with_routing_type3()
        self.test_json()

    def test_json_with_fragment(self):
        self.setUp_with_fragment()
        self.test_json()

    def test_json_with_auth(self):
        self.setUp_with_auth()
        self.test_json()

    def test_json_with_multi_headers(self):
        self.setUp_with_multi_headers()
        self.test_json()


class Test_hop_opts(unittest.TestCase):

    def setUp(self):
        self.nxt = 0
        self.size = 8
        self.data = [
            ipv6.option(5, 2, b'\x00\x00'),
            ipv6.option(1, 0, None),
            ipv6.option(0xc2, 4, b'\x00\x01\x00\x00'),
            ipv6.option(1, 0, None),
        ]
        self.hop = ipv6.hop_opts(self.nxt, self.size, self.data)
        self.form = '!BB'
        self.buf = struct.pack(self.form, self.nxt, self.size) \
            + self.data[0].serialize() \
            + self.data[1].serialize() \
            + self.data[2].serialize() \
            + self.data[3].serialize()

    def tearDown(self):
        pass

    def test_init(self):
        self.assertEqual(self.nxt, self.hop.nxt)
        self.assertEqual(self.size, self.hop.size)
        self.assertEqual(self.data, self.hop.data)

    def test_invalid_size(self):
        self.assertRaises(Exception, ipv6.hop_opts, self.nxt, 1, self.data)

    def test_parser(self):
        _res = ipv6.hop_opts.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        self.assertEqual(self.nxt, res.nxt)
        self.assertEqual(self.size, res.size)
        self.assertEqual(str(self.data), str(res.data))

    def test_serialize(self):
        buf = self.hop.serialize()
        res = struct.unpack_from(self.form, bytes(buf))
        self.assertEqual(self.nxt, res[0])
        self.assertEqual(self.size, res[1])
        offset = struct.calcsize(self.form)
        opt1 = ipv6.option.parser(bytes(buf[offset:]))
        offset += len(opt1)
        opt2 = ipv6.option.parser(bytes(buf[offset:]))
        offset += len(opt2)
        opt3 = ipv6.option.parser(bytes(buf[offset:]))
        offset += len(opt3)
        opt4 = ipv6.option.parser(bytes(buf[offset:]))
        self.assertEqual(5, opt1.type_)
        self.assertEqual(2, opt1.len_)
        self.assertEqual(b'\x00\x00', opt1.data)
        self.assertEqual(1, opt2.type_)
        self.assertEqual(0, opt2.len_)
        self.assertEqual(None, opt2.data)
        self.assertEqual(0xc2, opt3.type_)
        self.assertEqual(4, opt3.len_)
        self.assertEqual(b'\x00\x01\x00\x00', opt3.data)
        self.assertEqual(1, opt4.type_)
        self.assertEqual(0, opt4.len_)
        self.assertEqual(None, opt4.data)

    def test_len(self):
        self.assertEqual(16, len(self.hop))

    def test_default_args(self):
        hdr = ipv6.hop_opts()
        buf = hdr.serialize()
        res = struct.unpack('!BB', bytes(buf[:2]))

        self.assertEqual(res[0], 6)
        self.assertEqual(res[1], 0)
        opt = ipv6.option(type_=1, len_=4, data=b'\x00\x00\x00\x00')
        self.assertEqual(bytes(buf[2:]), opt.serialize())


class Test_dst_opts(unittest.TestCase):

    def setUp(self):
        self.nxt = 60
        self.size = 8
        self.data = [
            ipv6.option(5, 2, b'\x00\x00'),
            ipv6.option(1, 0, None),
            ipv6.option(0xc2, 4, b'\x00\x01\x00\x00'),
            ipv6.option(1, 0, None),
        ]
        self.dst = ipv6.dst_opts(self.nxt, self.size, self.data)
        self.form = '!BB'
        self.buf = struct.pack(self.form, self.nxt, self.size) \
            + self.data[0].serialize() \
            + self.data[1].serialize() \
            + self.data[2].serialize() \
            + self.data[3].serialize()

    def tearDown(self):
        pass

    def test_init(self):
        self.assertEqual(self.nxt, self.dst.nxt)
        self.assertEqual(self.size, self.dst.size)
        self.assertEqual(self.data, self.dst.data)

    def test_invalid_size(self):
        self.assertRaises(Exception, ipv6.dst_opts, self.nxt, 1, self.data)

    def test_parser(self):
        _res = ipv6.dst_opts.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        self.assertEqual(self.nxt, res.nxt)
        self.assertEqual(self.size, res.size)
        self.assertEqual(str(self.data), str(res.data))

    def test_serialize(self):
        buf = self.dst.serialize()
        res = struct.unpack_from(self.form, bytes(buf))
        self.assertEqual(self.nxt, res[0])
        self.assertEqual(self.size, res[1])
        offset = struct.calcsize(self.form)
        opt1 = ipv6.option.parser(bytes(buf[offset:]))
        offset += len(opt1)
        opt2 = ipv6.option.parser(bytes(buf[offset:]))
        offset += len(opt2)
        opt3 = ipv6.option.parser(bytes(buf[offset:]))
        offset += len(opt3)
        opt4 = ipv6.option.parser(bytes(buf[offset:]))
        self.assertEqual(5, opt1.type_)
        self.assertEqual(2, opt1.len_)
        self.assertEqual(b'\x00\x00', opt1.data)
        self.assertEqual(1, opt2.type_)
        self.assertEqual(0, opt2.len_)
        self.assertEqual(None, opt2.data)
        self.assertEqual(0xc2, opt3.type_)
        self.assertEqual(4, opt3.len_)
        self.assertEqual(b'\x00\x01\x00\x00', opt3.data)
        self.assertEqual(1, opt4.type_)
        self.assertEqual(0, opt4.len_)
        self.assertEqual(None, opt4.data)

    def test_len(self):
        self.assertEqual(16, len(self.dst))

    def test_default_args(self):
        hdr = ipv6.dst_opts()
        buf = hdr.serialize()
        res = struct.unpack('!BB', bytes(buf[:2]))

        self.assertEqual(res[0], 6)
        self.assertEqual(res[1], 0)
        opt = ipv6.option(type_=1, len_=4, data=b'\x00\x00\x00\x00')
        self.assertEqual(bytes(buf[2:]), opt.serialize())


class Test_option(unittest.TestCase):

    def setUp(self):
        self.type_ = 5
        self.data = b'\x00\x00'
        self.len_ = len(self.data)
        self.opt = ipv6.option(self.type_, self.len_, self.data)
        self.form = '!BB%ds' % self.len_
        self.buf = struct.pack(self.form, self.type_, self.len_, self.data)

    def tearDown(self):
        pass

    def test_init(self):
        self.assertEqual(self.type_, self.opt.type_)
        self.assertEqual(self.len_, self.opt.len_)
        self.assertEqual(self.data, self.opt.data)

    def test_parser(self):
        _res = ipv6.option.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        self.assertEqual(self.type_, res.type_)
        self.assertEqual(self.len_, res.len_)
        self.assertEqual(self.data, res.data)

    def test_serialize(self):
        buf = self.opt.serialize()
        res = struct.unpack_from(self.form, buf)
        self.assertEqual(self.type_, res[0])
        self.assertEqual(self.len_, res[1])
        self.assertEqual(self.data, res[2])

    def test_len(self):
        self.assertEqual(len(self.opt), 2 + self.len_)


class Test_option_pad1(Test_option):

    def setUp(self):
        self.type_ = 0
        self.len_ = -1
        self.data = None
        self.opt = ipv6.option(self.type_, self.len_, self.data)
        self.form = '!B'
        self.buf = struct.pack(self.form, self.type_)

    def test_serialize(self):
        buf = self.opt.serialize()
        res = struct.unpack_from(self.form, buf)
        self.assertEqual(self.type_, res[0])

    def test_default_args(self):
        opt = ipv6.option()
        buf = opt.serialize()
        res = struct.unpack('!B', buf)

        self.assertEqual(res[0], 0)


class Test_option_padN(Test_option):

    def setUp(self):
        self.type_ = 1
        self.len_ = 0
        self.data = None
        self.opt = ipv6.option(self.type_, self.len_, self.data)
        self.form = '!BB'
        self.buf = struct.pack(self.form, self.type_, self.len_)

    def test_serialize(self):
        buf = self.opt.serialize()
        res = struct.unpack_from(self.form, buf)
        self.assertEqual(self.type_, res[0])
        self.assertEqual(self.len_, res[1])


class Test_routing(unittest.TestCase):

    def setUp(self):
        self.nxt = 0
        self.size = 6
        self.type_ = ipv6.routing.ROUTING_TYPE_3
        self.seg = 0
        self.cmpi = 0
        self.cmpe = 0
        self.adrs = ["2001:db8:dead::1",
                     "2001:db8:dead::2",
                     "2001:db8:dead::3"]
        # calculate pad
        self.pad = (8 - ((len(self.adrs) - 1) * (16 - self.cmpi) +
                         (16 - self.cmpe) % 8)) % 8
        # create buf
        self.form = '!BBBBBB2x16s16s16s'
        self.buf = struct.pack(self.form, self.nxt, self.size,
                               self.type_, self.seg,
                               (self.cmpi << 4) | self.cmpe,
                               self.pad << 4,
                               addrconv.ipv6.text_to_bin(self.adrs[0]),
                               addrconv.ipv6.text_to_bin(self.adrs[1]),
                               addrconv.ipv6.text_to_bin(self.adrs[2]))

    def tearDown(self):
        pass

    def test_parser(self):
        _res = ipv6.routing.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        self.assertEqual(self.nxt, res.nxt)
        self.assertEqual(self.size, res.size)
        self.assertEqual(self.type_, res.type_)
        self.assertEqual(self.seg, res.seg)
        self.assertEqual(self.cmpi, res.cmpi)
        self.assertEqual(self.cmpe, res.cmpe)
        self.assertEqual(self.pad, res._pad)
        self.assertEqual(self.adrs[0], res.adrs[0])
        self.assertEqual(self.adrs[1], res.adrs[1])
        self.assertEqual(self.adrs[2], res.adrs[2])

    def test_not_implemented_type(self):
        not_implemented_buf = struct.pack(
            '!BBBBBB2x', 0, 6, ipv6.routing.ROUTING_TYPE_2, 0, 0, 0)
        instance = ipv6.routing.parser(not_implemented_buf)
        assert None is instance

    def test_invalid_type(self):
        invalid_type = 99
        invalid_buf = struct.pack('!BBBBBB2x', 0, 6, invalid_type, 0, 0, 0)
        instance = ipv6.routing.parser(invalid_buf)
        assert None is instance


class Test_routing_type3(unittest.TestCase):

    def setUp(self):
        self.nxt = 0
        self.size = 6
        self.type_ = 3
        self.seg = 0
        self.cmpi = 0
        self.cmpe = 0
        self.adrs = ["2001:db8:dead::1",
                     "2001:db8:dead::2",
                     "2001:db8:dead::3"]
        # calculate pad
        self.pad = (8 - ((len(self.adrs) - 1) * (16 - self.cmpi) +
                         (16 - self.cmpe) % 8)) % 8

        self.routing = ipv6.routing_type3(
            self.nxt, self.size, self.type_, self.seg, self.cmpi,
            self.cmpe, self.adrs)
        self.form = '!BBBBBB2x16s16s16s'
        self.buf = struct.pack(self.form, self.nxt, self.size,
                               self.type_, self.seg,
                               (self.cmpi << 4) | self.cmpe,
                               self.pad << 4,
                               addrconv.ipv6.text_to_bin(self.adrs[0]),
                               addrconv.ipv6.text_to_bin(self.adrs[1]),
                               addrconv.ipv6.text_to_bin(self.adrs[2]))

    def test_init(self):
        self.assertEqual(self.nxt, self.routing.nxt)
        self.assertEqual(self.size, self.routing.size)
        self.assertEqual(self.type_, self.routing.type_)
        self.assertEqual(self.seg, self.routing.seg)
        self.assertEqual(self.cmpi, self.routing.cmpi)
        self.assertEqual(self.cmpe, self.routing.cmpe)
        self.assertEqual(self.pad, self.routing._pad)
        self.assertEqual(self.adrs[0], self.routing.adrs[0])
        self.assertEqual(self.adrs[1], self.routing.adrs[1])
        self.assertEqual(self.adrs[2], self.routing.adrs[2])

    def test_parser(self):
        _res = ipv6.routing.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        self.assertEqual(self.nxt, res.nxt)
        self.assertEqual(self.size, res.size)
        self.assertEqual(self.type_, res.type_)
        self.assertEqual(self.seg, res.seg)
        self.assertEqual(self.cmpi, res.cmpi)
        self.assertEqual(self.cmpe, res.cmpe)
        self.assertEqual(self.pad, res._pad)
        self.assertEqual(self.adrs[0], res.adrs[0])
        self.assertEqual(self.adrs[1], res.adrs[1])
        self.assertEqual(self.adrs[2], res.adrs[2])

    def test_serialize(self):
        buf = self.routing.serialize()
        res = struct.unpack_from(self.form, bytes(buf))
        self.assertEqual(self.nxt, res[0])
        self.assertEqual(self.size, res[1])
        self.assertEqual(self.type_, res[2])
        self.assertEqual(self.seg, res[3])
        self.assertEqual(self.cmpi, res[4] >> 4)
        self.assertEqual(self.cmpe, res[4] & 0xf)
        self.assertEqual(self.pad, res[5])
        self.assertEqual(addrconv.ipv6.text_to_bin(self.adrs[0]), res[6])
        self.assertEqual(addrconv.ipv6.text_to_bin(self.adrs[1]), res[7])
        self.assertEqual(addrconv.ipv6.text_to_bin(self.adrs[2]), res[8])

    def test_parser_with_adrs_zero(self):
        nxt = 0
        size = 0
        type_ = 3
        seg = 0
        cmpi = 0
        cmpe = 0
        adrs = []
        # calculate pad
        pad = (8 - ((len(adrs) - 1) * (16 - cmpi) + (16 - cmpe) % 8)) % 8

        form = '!BBBBBB2x'
        buf = struct.pack(form, nxt, size, type_, seg,
                          (cmpi << 4) | cmpe, pad << 4)
        _res = ipv6.routing.parser(buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        self.assertEqual(nxt, res.nxt)
        self.assertEqual(size, res.size)
        self.assertEqual(type_, res.type_)
        self.assertEqual(seg, res.seg)
        self.assertEqual(cmpi, res.cmpi)
        self.assertEqual(cmpe, res.cmpe)
        self.assertEqual(pad, res._pad)

    def test_serialize_with_adrs_zero(self):
        nxt = 0
        size = 0
        type_ = 3
        seg = 0
        cmpi = 0
        cmpe = 0
        adrs = []
        # calculate pad
        pad = (8 - ((len(adrs) - 1) * (16 - cmpi) + (16 - cmpe) % 8)) % 8
        routing = ipv6.routing_type3(
            nxt, size, type_, seg, cmpi,
            cmpe, pad)
        buf = routing.serialize()
        form = '!BBBBBB2x'
        res = struct.unpack_from(form, bytes(buf))
        self.assertEqual(nxt, res[0])
        self.assertEqual(size, res[1])
        self.assertEqual(type_, res[2])
        self.assertEqual(seg, res[3])
        self.assertEqual(cmpi, res[4] >> 4)
        self.assertEqual(cmpe, res[4] & 0xf)
        self.assertEqual(pad, res[5])

    def test_parser_with_compression(self):
        pass
        nxt = 0
        size = 3
        type_ = 3
        seg = 0
        cmpi = 8
        cmpe = 12
        adrs = ["2001:0db8:dead:0123:4567:89ab:cdef:0001",
                "2001:0db8:dead:0123:4567:89ab:cdef:0002",
                "2001:0db8:dead:0123:4567:89ab:cdef:0003"]
        # calculate pad
        pad = (8 - ((len(adrs) - 1) * (16 - cmpi) + (16 - cmpe) % 8)) % 8
        form = '!BBBBBB2x%ds%ds%ds' % (16 - cmpi, 16 - cmpi, 16 - cmpe)
        slice_i = slice(cmpi, 16)
        slice_e = slice(cmpe, 16)
        buf = struct.pack(form, nxt, size, type_, seg,
                          (cmpi << 4) | cmpe, pad << 4,
                          addrconv.ipv6.text_to_bin(adrs[0])[slice_i],
                          addrconv.ipv6.text_to_bin(adrs[1])[slice_i],
                          addrconv.ipv6.text_to_bin(adrs[2])[slice_e])
        _res = ipv6.routing.parser(buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        self.assertEqual(nxt, res.nxt)
        self.assertEqual(size, res.size)
        self.assertEqual(type_, res.type_)
        self.assertEqual(seg, res.seg)
        self.assertEqual(cmpi, res.cmpi)
        self.assertEqual(cmpe, res.cmpe)
        self.assertEqual(pad, res._pad)
        self.assertEqual("::4567:89ab:cdef:1", res.adrs[0])
        self.assertEqual("::4567:89ab:cdef:2", res.adrs[1])
        self.assertEqual("::205.239.0.3", res.adrs[2])

    def test_serialize_with_compression(self):
        nxt = 0
        size = 3
        type_ = 3
        seg = 0
        cmpi = 8
        cmpe = 8
        adrs = ["2001:db8:dead::1",
                "2001:db8:dead::2",
                "2001:db8:dead::3"]
        # calculate pad
        pad = (8 - ((len(adrs) - 1) * (16 - cmpi) + (16 - cmpe) % 8)) % 8
        slice_i = slice(cmpi, 16)
        slice_e = slice(cmpe, 16)
        routing = ipv6.routing_type3(
            nxt, size, type_, seg, cmpi, cmpe, adrs)
        buf = routing.serialize()
        form = '!BBBBBB2x8s8s8s'
        res = struct.unpack_from(form, bytes(buf))
        self.assertEqual(nxt, res[0])
        self.assertEqual(size, res[1])
        self.assertEqual(type_, res[2])
        self.assertEqual(seg, res[3])
        self.assertEqual(cmpi, res[4] >> 4)
        self.assertEqual(cmpe, res[4] & 0xf)
        self.assertEqual(pad, res[5])
        self.assertEqual(addrconv.ipv6.text_to_bin(adrs[0])[slice_i], res[6])
        self.assertEqual(addrconv.ipv6.text_to_bin(adrs[1])[slice_i], res[7])
        self.assertEqual(addrconv.ipv6.text_to_bin(adrs[2])[slice_e], res[8])

    def test_len(self):
        self.assertEqual((6 + 1) * 8, len(self.routing))

    def test_default_args(self):
        hdr = ipv6.routing_type3()
        buf = hdr.serialize()
        LOG.info(repr(buf))
        res = struct.unpack_from(ipv6.routing_type3._PACK_STR, bytes(buf))
        LOG.info(res)

        self.assertEqual(res[0], 6)
        self.assertEqual(res[1], 0)
        self.assertEqual(res[2], 3)
        self.assertEqual(res[3], 0)
        self.assertEqual(res[4], (0 << 4) | 0)
        self.assertEqual(res[5], 0)


class Test_fragment(unittest.TestCase):

    def setUp(self):
        self.nxt = 44
        self.offset = 50
        self.more = 1
        self.id_ = 123
        self.fragment = ipv6.fragment(
            self.nxt, self.offset, self.more, self.id_)

        self.off_m = (self.offset << 3 | self.more)
        self.form = '!BxHI'
        self.buf = struct.pack(self.form, self.nxt, self.off_m, self.id_)

    def test_init(self):
        self.assertEqual(self.nxt, self.fragment.nxt)
        self.assertEqual(self.offset, self.fragment.offset)
        self.assertEqual(self.more, self.fragment.more)
        self.assertEqual(self.id_, self.fragment.id_)

    def test_parser(self):
        _res = ipv6.fragment.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        self.assertEqual(self.nxt, res.nxt)
        self.assertEqual(self.offset, res.offset)
        self.assertEqual(self.more, res.more)
        self.assertEqual(self.id_, res.id_)

    def test_serialize(self):
        buf = self.fragment.serialize()
        res = struct.unpack_from(self.form, bytes(buf))
        self.assertEqual(self.nxt, res[0])
        self.assertEqual(self.off_m, res[1])
        self.assertEqual(self.id_, res[2])

    def test_len(self):
        self.assertEqual(8, len(self.fragment))

    def test_default_args(self):
        hdr = ipv6.fragment()
        buf = hdr.serialize()
        res = struct.unpack_from(ipv6.fragment._PACK_STR, buf)

        self.assertEqual(res[0], 6)
        self.assertEqual(res[1], 0)
        self.assertEqual(res[2], 0)


class Test_auth(unittest.TestCase):

    def setUp(self):
        self.nxt = 0
        self.size = 4
        self.spi = 256
        self.seq = 1
        self.data = b'\x21\xd3\xa9\x5c\x5f\xfd\x4d\x18\x46\x22\xb9\xf8'
        self.auth = ipv6.auth(
            self.nxt, self.size, self.spi, self.seq, self.data)
        self.form = '!BB2xII12s'
        self.buf = struct.pack(self.form, self.nxt, self.size, self.spi,
                               self.seq, self.data)

    def test_init(self):
        self.assertEqual(self.nxt, self.auth.nxt)
        self.assertEqual(self.size, self.auth.size)
        self.assertEqual(self.spi, self.auth.spi)
        self.assertEqual(self.seq, self.auth.seq)
        self.assertEqual(self.data, self.auth.data)

    def test_parser(self):
        _res = ipv6.auth.parser(self.buf)
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        self.assertEqual(self.nxt, res.nxt)
        self.assertEqual(self.size, res.size)
        self.assertEqual(self.spi, res.spi)
        self.assertEqual(self.seq, res.seq)
        self.assertEqual(self.data, res.data)

    def test_serialize(self):
        buf = self.auth.serialize()
        res = struct.unpack_from(self.form, bytes(buf))
        self.assertEqual(self.nxt, res[0])
        self.assertEqual(self.size, res[1])
        self.assertEqual(self.spi, res[2])
        self.assertEqual(self.seq, res[3])
        self.assertEqual(self.data, res[4])

    def test_len(self):
        self.assertEqual((4 + 2) * 4, len(self.auth))

    def test_len_re(self):
        size = 5
        auth = ipv6.auth(
            0, size, 256, 1,
            b'\x21\xd3\xa9\x5c\x5f\xfd\x4d\x18\x46\x22\xb9\xf8\xf8\xf8\xf8\xf8')
        self.assertEqual((size + 2) * 4, len(auth))

    def test_default_args(self):
        hdr = ipv6.auth()
        buf = hdr.serialize()
        LOG.info(repr(buf))
        res = struct.unpack_from(ipv6.auth._PACK_STR, bytes(buf))
        LOG.info(res)

        self.assertEqual(res[0], 6)
        self.assertEqual(res[1], 2)
        self.assertEqual(res[2], 0)
        self.assertEqual(res[3], 0)
        self.assertEqual(buf[ipv6.auth._MIN_LEN:], b'\x00\x00\x00\x00')



# Copyright (C) 2015 Nippon Telegraph and Telephone Corporation.
#
# This is based on the following
#     https://github.com/osrg/gobgp/test/lib/base.py
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import logging
import os
import re
import subprocess
import time

import netaddr


LOG = logging.getLogger(__name__)

DEFAULT_TEST_PREFIX = ''
DEFAULT_TEST_BASE_DIR = '/tmp/ctn_docker/bgp'
TEST_PREFIX = DEFAULT_TEST_PREFIX
TEST_BASE_DIR = DEFAULT_TEST_BASE_DIR

BGP_FSM_IDLE = 'BGP_FSM_IDLE'
BGP_FSM_ACTIVE = 'BGP_FSM_ACTIVE'
BGP_FSM_ESTABLISHED = 'BGP_FSM_ESTABLISHED'

BGP_ATTR_TYPE_ORIGIN = 1
BGP_ATTR_TYPE_AS_PATH = 2
BGP_ATTR_TYPE_NEXT_HOP = 3
BGP_ATTR_TYPE_MULTI_EXIT_DISC = 4
BGP_ATTR_TYPE_LOCAL_PREF = 5
BGP_ATTR_TYPE_COMMUNITIES = 8
BGP_ATTR_TYPE_ORIGINATOR_ID = 9
BGP_ATTR_TYPE_CLUSTER_LIST = 10
BGP_ATTR_TYPE_MP_REACH_NLRI = 14
BGP_ATTR_TYPE_EXTENDED_COMMUNITIES = 16

BRIDGE_TYPE_DOCKER = 'docker'
BRIDGE_TYPE_BRCTL = 'brctl'
BRIDGE_TYPE_OVS = 'ovs'


class CommandError(Exception):
    def __init__(self, out):
        super(CommandError, self).__init__()
        self.out = out


def try_several_times(f, t=3, s=1):
    e = RuntimeError()
    for _ in range(t):
        try:
            r = f()
        except RuntimeError as e:
            time.sleep(s)
        else:
            return r
    raise e


class CmdBuffer(list):
    def __init__(self, delim='\n'):
        super(CmdBuffer, self).__init__()
        self.delim = delim

    def __lshift__(self, value):
        self.append(value)

    def __str__(self):
        return self.delim.join(self)


class CommandOut(str):

    def __new__(cls, stdout, stderr, command, returncode, **kwargs):
        stdout = stdout or ''
        obj = super(CommandOut, cls).__new__(cls, stdout, **kwargs)
        obj.stderr = stderr or ''
        obj.command = command
        obj.returncode = returncode
        return obj


class Command(object):

    def _execute(self, cmd, capture=False, executable=None):
        """Execute a command using subprocess.Popen()
        :Parameters:
            - out: stdout from subprocess.Popen()
              out has some attributes.
              out.returncode: returncode of subprocess.Popen()
              out.stderr: stderr from subprocess.Popen()
        """
        if capture:
            p_stdout = subprocess.PIPE
            p_stderr = subprocess.PIPE
        else:
            p_stdout = None
            p_stderr = None
        pop = subprocess.Popen(cmd, shell=True, executable=executable,
                               stdout=p_stdout,
                               stderr=p_stderr)
        __stdout, __stderr = pop.communicate()
        _stdout = str(__stdout, 'utf-8')
        _stderr = str(__stderr, 'utf-8')
        out = CommandOut(_stdout, _stderr, cmd, pop.returncode)
        return out

    def execute(self, cmd, capture=True, try_times=1, interval=1):
        out = None
        for i in range(try_times):
            out = self._execute(cmd, capture=capture)
            LOG.info(out.command)
            if out.returncode == 0:
                return out
            LOG.error("stdout: %s", out)
            LOG.error("stderr: %s", out.stderr)
            if i + 1 >= try_times:
                break
            time.sleep(interval)
        raise CommandError(out)

    def sudo(self, cmd, capture=True, try_times=1, interval=1):
        cmd = 'sudo %s' % cmd
        return self.execute(cmd, capture=capture,
                            try_times=try_times, interval=interval)


class DockerImage(object):
    def __init__(self, baseimage='ubuntu:20.04'):
        self.baseimage = baseimage
        self.cmd = Command()

    def get_images(self):
        out = self.cmd.sudo('docker images')
        images = []
        for line in out.splitlines()[1:]:
            images.append(line.split()[0])
        return images

    def exist(self, name):
        return name in self.get_images()

    def build(self, tagname, dockerfile_dir):
        try:
            self.cmd.sudo(
                'docker build -t {0} {1}'.format(tagname, dockerfile_dir))
        except CommandError:
            self.cmd.sudo('docker builder prune -f')
            self.cmd.sudo(
                'docker build -t {0} {1}'.format(tagname, dockerfile_dir),
                try_times=2)

    def remove(self, tagname, check_exist=False):
        if check_exist and not self.exist(tagname):
            return tagname
        self.cmd.sudo("docker rmi -f %s" % tagname, try_times=3)

    def create_quagga(self, tagname='quagga', image=None, check_exist=False):
        if check_exist and self.exist(tagname):
            return tagname
        workdir = os.path.join(TEST_BASE_DIR, tagname)
        pkges = ' '.join([
            'telnet',
            'tcpdump',
            'quagga-bgpd',
        ])
        if image:
            use_image = image
        else:
            use_image = self.baseimage
        c = CmdBuffer()
        c << 'FROM %s' % use_image
        c << 'RUN apt-get update'
        c << 'RUN apt-get install -qy --no-install-recommends %s' % pkges
        c << 'RUN echo "#!/bin/sh" > /bgpd'
        c << 'RUN echo mkdir -p /run/quagga >> /bgpd'
        c << 'RUN echo chmod 755 /run/quagga >> /bgpd'
        c << 'RUN echo chown quagga:quagga /run/quagga >> /bgpd'
        c << 'RUN echo exec /usr/sbin/bgpd >> /bgpd'
        c << 'RUN chmod +x /bgpd'
        c << 'CMD /bgpd'

        self.cmd.sudo('rm -rf %s' % workdir)
        self.cmd.execute('mkdir -p %s' % workdir)
        self.cmd.execute("echo '%s' > %s/Dockerfile" % (str(c), workdir))
        self.build(tagname, workdir)
        return tagname

    def create_os_ken(self, tagname='os_ken', image=None, check_exist=False):
        if check_exist and self.exist(tagname):
            return tagname
        workdir = os.path.join(TEST_BASE_DIR, tagname)
        workdir_ctn = '/root/osrg/os_ken'
        pkges = ' '.join([
            'tcpdump',
            'iproute2',
        ])
        if image:
            use_image = image
        else:
            use_image = self.baseimage
        c = CmdBuffer()
        c << 'FROM %s' % use_image
        c << 'ADD os_ken %s' % workdir_ctn
        install = ' '.join([
            'RUN apt-get update',
            '&& apt-get install -qy --no-install-recommends %s' % pkges,
            '&& cd %s' % workdir_ctn,
            # Note: Clean previous builds, because "python setup.py install"
            # might fail if the current directory contains the symlink to
            # Docker host file systems.
            '&& rm -rf *.egg-info/ build/ dist/ .tox/ *.log'
            '&& pip install -r requirements.txt -r test-requirements.txt',
            '&& pip install .',
        ])
        c << install

        self.cmd.sudo('rm -rf %s' % workdir)
        self.cmd.execute('mkdir -p %s' % workdir)
        self.cmd.execute("echo '%s' > %s/Dockerfile" % (str(c), workdir))
        self.cmd.execute('cp -r ../os_ken %s/' % workdir)
        self.build(tagname, workdir)
        return tagname


class Bridge(object):
    def __init__(self, name, subnet='', start_ip=None, end_ip=None,
                 with_ip=True, self_ip=False,
                 fixed_ip=None, reuse=False,
                 br_type='docker'):
        """Manage a bridge
        :Parameters:
            - name: bridge name
            - subnet: network cider to be used in this bridge
            - start_ip: start address of an ip to be used in the subnet
            - end_ip: end address of an ip to be used in the subnet
            - with_ip: specify if assign automatically an ip address
            - self_ip: specify if assign an ip address for the bridge
            - fixed_ip: an ip address to be assigned to the bridge
            - reuse: specify if use an existing bridge
            - br_type: One either in a 'docker', 'brctl' or 'ovs'
        """
        self.cmd = Command()
        self.name = name
        if br_type not in (BRIDGE_TYPE_DOCKER, BRIDGE_TYPE_BRCTL,
                           BRIDGE_TYPE_OVS):
            raise Exception("argument error br_type: %s" % br_type)
        self.br_type = br_type
        self.docker_nw = bool(self.br_type == BRIDGE_TYPE_DOCKER)
        if TEST_PREFIX != '':
            self.name = '{0}_{1}'.format(TEST_PREFIX, name)
        self.with_ip = with_ip
        if with_ip:
            self.subnet = netaddr.IPNetwork(subnet)
            if start_ip:
                self.start_ip = start_ip
            else:
                self.start_ip = netaddr.IPAddress(self.subnet.first)
            if end_ip:
                self.end_ip = end_ip
            else:
                self.end_ip = netaddr.IPAddress(self.subnet.last)

            def _ip_gen():
                for host in netaddr.IPRange(self.start_ip, self.end_ip):
                    yield host
            self._ip_generator = _ip_gen()
            # throw away first network address
            self.next_ip_address()

        self.self_ip = self_ip
        if fixed_ip:
            self.ip_addr = fixed_ip
        else:
            self.ip_addr = self.next_ip_address()
        if not reuse:
            def f():
                if self.br_type == BRIDGE_TYPE_DOCKER:
                    gw = "--gateway %s" % self.ip_addr.split('/')[0]
                    v6 = ''
                    if self.subnet.version == 6:
                        v6 = '--ipv6'
                    cmd = ("docker network create --driver bridge %s "
                           "%s --subnet %s %s" % (v6, gw, subnet, self.name))
                elif self.br_type == BRIDGE_TYPE_BRCTL:
                    cmd = "ip link add {0} type bridge".format(self.name)
                elif self.br_type == BRIDGE_TYPE_OVS:
                    cmd = "ovs-vsctl add-br {0}".format(self.name)
                else:
                    raise ValueError('Unsupported br_type: %s' % self.br_type)
                self.delete()
                self.execute(cmd, sudo=True, retry=True)
            try_several_times(f)
        if not self.docker_nw:
            self.execute("ip link set up dev {0}".format(self.name),
                         sudo=True, retry=True)

        if not self.docker_nw and self_ip:
            ips = self.check_br_addr(self.name)
            for key, ip in ips.items():
                if self.subnet.version == key:
                    self.execute(
                        "ip addr del {0} dev {1}".format(ip, self.name),
                        sudo=True, retry=True)
            self.execute(
                "ip addr add {0} dev {1}".format(self.ip_addr, self.name),
                sudo=True, retry=True)
        self.ctns = []

    def get_bridges_dc(self):
        out = self.execute('docker network ls', sudo=True, retry=True)
        bridges = []
        for line in out.splitlines()[1:]:
            bridges.append(line.split()[1])
        return bridges

    def get_bridges_brctl(self):
        br_list = []
        bridges = glob.glob('/sys/class/net/*/bridge/bridge_id')
        regex = re.compile(r"\/sys\/class\/net\/(.+)\/bridge\/bridge_id")
        for bridge in bridges:
            m = regex.match(bridge)
            br_list.append(m.group(1))
        return br_list

    def get_bridges_ovs(self):
        out = self.execute('ovs-vsctl list-br', sudo=True, retry=True)
        return out.splitlines()

    def get_bridges(self):
        if self.br_type == BRIDGE_TYPE_DOCKER:
            return self.get_bridges_dc()
        elif self.br_type == BRIDGE_TYPE_BRCTL:
            return self.get_bridges_brctl()
        elif self.br_type == BRIDGE_TYPE_OVS:
            return self.get_bridges_ovs()

    def exist(self):
        return self.name in self.get_bridges()

    def execute(self, cmd, capture=True, sudo=False, retry=False):
        if sudo:
            m = self.cmd.sudo
        else:
            m = self.cmd.execute
        if retry:
            return m(cmd, capture=capture, try_times=3, interval=1)
        else:
            return m(cmd, capture=capture)

    def check_br_addr(self, br):
        ips = {}
        cmd = "ip a show dev %s" % br
        for line in self.execute(cmd, sudo=True).split('\n'):
            if line.strip().startswith("inet "):
                elems = [e.strip() for e in line.strip().split(' ')]
                ips[4] = elems[1]
            elif line.strip().startswith("inet6 "):
                elems = [e.strip() for e in line.strip().split(' ')]
                ips[6] = elems[1]
        return ips

    def next_ip_address(self):
        return "{0}/{1}".format(next(self._ip_generator),
                                self.subnet.prefixlen)

    def addif(self, ctn):
        name = ctn.next_if_name()
        self.ctns.append(ctn)
        ip_address = None
        if self.docker_nw:
            ipv4 = None
            ipv6 = None
            ip_address = self.next_ip_address()
            ip_address_ip = ip_address.split('/')[0]
            version = 4
            if netaddr.IPNetwork(ip_address).version == 6:
                version = 6
            opt_ip = "--ip %s" % ip_address_ip
            if version == 4:
                ipv4 = ip_address
            else:
                opt_ip = "--ip6 %s" % ip_address_ip
                ipv6 = ip_address
            cmd = "docker network connect %s %s %s" % (
                opt_ip, self.name, ctn.docker_name())
            self.execute(cmd, sudo=True)
            ctn.set_addr_info(bridge=self.name, ipv4=ipv4, ipv6=ipv6,
                              ifname=name)
        else:
            if self.with_ip:
                ip_address = self.next_ip_address()
                version = 4
                if netaddr.IPNetwork(ip_address).version == 6:
                    version = 6
                ctn.pipework(self, ip_address, name, version=version)
            else:
                ctn.pipework(self, '0/0', name)
        return ip_address

    def delete(self, check_exist=True):
        if check_exist:
            if not self.exist():
                return
        if self.br_type == BRIDGE_TYPE_DOCKER:
            self.execute("docker network rm %s" % self.name,
                         sudo=True, retry=True)
        elif self.br_type == BRIDGE_TYPE_BRCTL:
            self.execute("ip link set down dev %s" % self.name,
                         sudo=True, retry=True)
            self.execute(
                "ip link delete %s type bridge" % self.name,
                sudo=True, retry=True)
        elif self.br_type == BRIDGE_TYPE_OVS:
            self.execute(
                "ovs-vsctl del-br %s" % self.name,
                sudo=True, retry=True)


class Container(object):
    def __init__(self, name, image=None):
        self.name = name
        self.image = image
        self.shared_volumes = []
        self.ip_addrs = []
        self.ip6_addrs = []
        self.is_running = False
        self.eths = []
        self.id = None

        self.cmd = Command()
        self.remove()

    def docker_name(self):
        if TEST_PREFIX == DEFAULT_TEST_PREFIX:
            return self.name
        return '{0}_{1}'.format(TEST_PREFIX, self.name)

    def get_docker_id(self):
        if self.id:
            return self.id
        else:
            return self.docker_name()

    def next_if_name(self):
        name = 'eth{0}'.format(len(self.eths) + 1)
        self.eths.append(name)
        return name

    def set_addr_info(self, bridge, ipv4=None, ipv6=None, ifname='eth0'):
        if ipv4:
            self.ip_addrs.append((ifname, ipv4, bridge))
        if ipv6:
            self.ip6_addrs.append((ifname, ipv6, bridge))

    def get_addr_info(self, bridge, ipv=4):
        addrinfo = {}
        if ipv == 4:
            ip_addrs = self.ip_addrs
        elif ipv == 6:
            ip_addrs = self.ip6_addrs
        else:
            return None
        for addr in ip_addrs:
            if addr[2] == bridge:
                addrinfo[addr[1]] = addr[0]
        return addrinfo

    def execute(self, cmd, capture=True, sudo=False, retry=False):
        if sudo:
            m = self.cmd.sudo
        else:
            m = self.cmd.execute
        if retry:
            return m(cmd, capture=capture, try_times=3, interval=1)
        else:
            return m(cmd, capture=capture)

    def dcexec(self, cmd, capture=True, retry=False):
        if retry:
            return self.cmd.sudo(cmd, capture=capture, try_times=3, interval=1)
        else:
            return self.cmd.sudo(cmd, capture=capture)

    def exec_on_ctn(self, cmd, capture=True, detach=False):
        name = self.docker_name()
        flag = '-d' if detach else ''
        return self.dcexec('docker exec {0} {1} {2}'.format(
            flag, name, cmd), capture=capture)

    def get_containers(self, allctn=False):
        cmd = 'docker ps --no-trunc=true'
        if allctn:
            cmd += ' --all=true'
        out = self.dcexec(cmd, retry=True)
        containers = []
        for line in out.splitlines()[1:]:
            containers.append(line.split()[-1])
        return containers

    def exist(self, allctn=False):
        return self.docker_name() in self.get_containers(allctn=allctn)

    def run(self):
        c = CmdBuffer(' ')
        c << "docker run --privileged=true"
        for sv in self.shared_volumes:
            c << "-v {0}:{1}".format(sv[0], sv[1])
        c << "--name {0} --hostname {0} -id {1}".format(self.docker_name(),
                                                        self.image)
        self.id = self.dcexec(str(c), retry=True)
        self.is_running = True
        self.exec_on_ctn("ip li set up dev lo")
        ipv4 = None
        ipv6 = None
        for line in self.exec_on_ctn("ip a show dev eth0").split('\n'):
            if line.strip().startswith("inet "):
                elems = [e.strip() for e in line.strip().split(' ')]
                ipv4 = elems[1]
            elif line.strip().startswith("inet6 "):
                elems = [e.strip() for e in line.strip().split(' ')]
                ipv6 = elems[1]
        self.set_addr_info(bridge='docker0', ipv4=ipv4, ipv6=ipv6,
                           ifname='eth0')
        return 0

    def stop(self, check_exist=True):
        if check_exist:
            if not self.exist(allctn=False):
                return
        ctn_id = self.get_docker_id()
        out = self.dcexec('docker stop -t 0 %s' % ctn_id, retry=True)
        self.is_running = False
        return out

    def remove(self, check_exist=True):
        if check_exist:
            if not self.exist(allctn=True):
                return
        ctn_id = self.get_docker_id()
        out = self.dcexec('docker rm -f %s' % ctn_id, retry=True)
        self.is_running = False
        return out

    def pipework(self, bridge, ip_addr, intf_name="", version=4):
        if not self.is_running:
            LOG.warning('Call run() before pipeworking')
            return
        c = CmdBuffer(' ')
        c << "pipework {0}".format(bridge.name)

        if intf_name != "":
            c << "-i {0}".format(intf_name)
        else:
            intf_name = "eth1"
        ipv4 = None
        ipv6 = None
        if version == 4:
            ipv4 = ip_addr
        else:
            c << '-a 6'
            ipv6 = ip_addr
        c << "{0} {1}".format(self.docker_name(), ip_addr)
        self.set_addr_info(bridge=bridge.name, ipv4=ipv4, ipv6=ipv6,
                           ifname=intf_name)
        self.execute(str(c), sudo=True, retry=True)

    def get_pid(self):
        if self.is_running:
            cmd = "docker inspect -f '{{.State.Pid}}' %s" % self.docker_name()
            return int(self.dcexec(cmd))
        return -1

    def start_tcpdump(self, interface=None, filename=None):
        if not interface:
            interface = "eth0"
        if not filename:
            filename = "{0}/{1}.dump".format(
                self.shared_volumes[0][1], interface)
        self.exec_on_ctn(
            "tcpdump -i {0} -w {1}".format(interface, filename),
            detach=True)


class BGPContainer(Container):

    WAIT_FOR_BOOT = 1
    RETRY_INTERVAL = 5
    DEFAULT_PEER_ARG = {'neigh_addr': '',
                        'passwd': None,
                        'vpn': False,
                        'flowspec': False,
                        'is_rs_client': False,
                        'is_rr_client': False,
                        'cluster_id': None,
                        'policies': None,
                        'passive': False,
                        'local_addr': '',
                        'as2': False,
                        'graceful_restart': None,
                        'local_as': None,
                        'prefix_limit': None}
    default_peer_keys = sorted(DEFAULT_PEER_ARG.keys())
    DEFAULT_ROUTE_ARG = {'prefix': None,
                         'rf': 'ipv4',
                         'attr': None,
                         'next-hop': None,
                         'as-path': None,
                         'community': None,
                         'med': None,
                         'local-pref': None,
                         'extended-community': None,
                         'matchs': None,
                         'thens': None}
    default_route_keys = sorted(DEFAULT_ROUTE_ARG.keys())

    def __init__(self, name, asn, router_id, ctn_image_name=None):
        self.config_dir = TEST_BASE_DIR
        if TEST_PREFIX:
            self.config_dir = os.path.join(self.config_dir, TEST_PREFIX)
        self.config_dir = os.path.join(self.config_dir, name)
        self.asn = asn
        self.router_id = router_id
        self.peers = {}
        self.routes = {}
        self.policies = {}
        super(BGPContainer, self).__init__(name, ctn_image_name)
        self.execute(
            'rm -rf {0}'.format(self.config_dir), sudo=True)
        self.execute('mkdir -p {0}'.format(self.config_dir))
        self.execute('chmod 777 {0}'.format(self.config_dir))

    def __repr__(self):
        return str({'name': self.name, 'asn': self.asn,
                    'router_id': self.router_id})

    def run(self, wait=False, w_time=WAIT_FOR_BOOT):
        self.create_config()
        super(BGPContainer, self).run()
        if wait:
            time.sleep(w_time)
        return w_time

    def add_peer(self, peer, bridge='', reload_config=True, v6=False,
                 peer_info=None):
        peer_info = peer_info or {}
        self.peers[peer] = self.DEFAULT_PEER_ARG.copy()
        self.peers[peer].update(peer_info)
        peer_keys = sorted(self.peers[peer].keys())
        if peer_keys != self.default_peer_keys:
            raise Exception("argument error peer_info: %s" % peer_info)

        neigh_addr = ''
        local_addr = ''
        it = itertools.product(self.ip_addrs, peer.ip_addrs)
        if v6:
            it = itertools.product(self.ip6_addrs, peer.ip6_addrs)

        for me, you in it:
            if bridge != '' and bridge != me[2]:
                continue
            if me[2] == you[2]:
                neigh_addr = you[1]
                local_addr = me[1]
                if v6:
                    addr, mask = local_addr.split('/')
                    local_addr = "{0}%{1}/{2}".format(addr, me[0], mask)
                break

        if neigh_addr == '':
            raise Exception('peer {0} seems not ip reachable'.format(peer))

        if not self.peers[peer]['policies']:
            self.peers[peer]['policies'] = {}

        self.peers[peer]['neigh_addr'] = neigh_addr
        self.peers[peer]['local_addr'] = local_addr
        if self.is_running and reload_config:
            self.create_config()
            self.reload_config()

    def del_peer(self, peer, reload_config=True):
        del self.peers[peer]
        if self.is_running and reload_config:
            self.create_config()
            self.reload_config()

    def disable_peer(self, peer):
        raise NotImplementedError()

    def enable_peer(self, peer):
        raise NotImplementedError()

    def log(self):
        return self.execute('cat {0}/*.log'.format(self.config_dir))

    def add_route(self, route, reload_config=True, route_info=None):
        route_info = route_info or {}
        self.routes[route] = self.DEFAULT_ROUTE_ARG.copy()
        self.routes[route].update(route_info)
        route_keys = sorted(self.routes[route].keys())
        if route_keys != self.default_route_keys:
            raise Exception("argument error route_info: %s" % route_info)
        self.routes[route]['prefix'] = route
        if self.is_running and reload_config:
            self.create_config()
            self.reload_config()

    def add_policy(self, policy, peer, typ, default='accept',
                   reload_config=True):
        self.set_default_policy(peer, typ, default)
        self.define_policy(policy)
        self.assign_policy(peer, policy, typ)
        if self.is_running and reload_config:
            self.create_config()
            self.reload_config()

    def set_default_policy(self, peer, typ, default):
        if (typ in ['in', 'out', 'import', 'export'] and
                default in ['reject', 'accept']):
            if 'default-policy' not in self.peers[peer]:
                self.peers[peer]['default-policy'] = {}
            self.peers[peer]['default-policy'][typ] = default
        else:
            raise Exception('wrong type or default')

    def define_policy(self, policy):
        self.policies[policy['name']] = policy

    def assign_policy(self, peer, policy, typ):
        if peer not in self.peers:
            raise Exception('peer {0} not found'.format(peer.name))
        name = policy['name']
        if name not in self.policies:
            raise Exception('policy {0} not found'.format(name))
        self.peers[peer]['policies'][typ] = policy

    def get_local_rib(self, peer, rf):
        raise NotImplementedError()

    def get_global_rib(self, rf):
        raise NotImplementedError()

    def get_neighbor_state(self, peer_id):
        raise NotImplementedError()

    def get_reachablily(self, prefix, timeout=20):
        version = netaddr.IPNetwork(prefix).version
        addr = prefix.split('/')[0]
        if version == 4:
            ping_cmd = 'ping'
        elif version == 6:
            ping_cmd = 'ping6'
        else:
            raise Exception(
                'unsupported route family: {0}'.format(version))
        cmd = '/bin/bash -c "/bin/{0} -c 1 -w 1 {1} | xargs echo"'.format(
            ping_cmd, addr)
        interval = 1
        count = 0
        while True:
            res = self.exec_on_ctn(cmd)
            LOG.info(res)
            if '1 packets received' in res and '0% packet loss':
                break
            time.sleep(interval)
            count += interval
            if count >= timeout:
                raise Exception('timeout')
        return True

    def wait_for(self, expected_state, peer, timeout=120):
        interval = 1
        count = 0
        while True:
            state = self.get_neighbor_state(peer)
            LOG.info("%s's peer %s state: %s",
                     self.router_id, peer.router_id, state)
            if state == expected_state:
                return

            time.sleep(interval)
            count += interval
            if count >= timeout:
                raise Exception('timeout')

    def add_static_route(self, network, next_hop):
        cmd = '/sbin/ip route add {0} via {1}'.format(network, next_hop)
        self.exec_on_ctn(cmd)

    def set_ipv6_forward(self):
        cmd = 'sysctl -w net.ipv6.conf.all.forwarding=1'
        self.exec_on_ctn(cmd)

    def create_config(self):
        raise NotImplementedError()

    def reload_config(self):
        raise NotImplementedError()



# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2013 Isaku Yamahata <yamahata at private email ne jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
OF-Config implementation.
"""

import glob
import os.path
import sys


SCHEMA_DIR = os.path.dirname(__file__)

_PREFIX = 'of-config-'
_SUFFIX = '.xsd'
_files = glob.glob(os.path.join(SCHEMA_DIR, 'of-config-*.xsd'))
OF_CONFIG_XSD_FILES = dict(
    (os.path.basename(f)[len(_PREFIX):-len(_SUFFIX)], f) for f in _files)

# For convenience
# OF_CONFIG_1_0_XSD = os.path.join(SCHEMA_DIR, 'of-config-1.0.xsd')
# and so on
_this_module = sys.modules[__name__]
for (version, xsd_file) in OF_CONFIG_XSD_FILES.items():
    setattr(_this_module,
            'OF_CONFIG_%s_XSD' % version.replace('.', '_'), xsd_file)


OFCONFIG_1_1_CONFIG = 'urn:onf:params:xml:ns:onf:of12:config'
OFCONFIG_1_1_YANG = 'urn:onf:of12:config:yang'

# LINC specific?
OFCONFIG_1_1_1_YANG = 'urn:onf:of111:config:yang'

OFCONFIG_YANG_NAMESPACES = {
    '1.1': OFCONFIG_1_1_YANG,
    '1.1.1': OFCONFIG_1_1_1_YANG,
}



import logging
LOG = logging.getLogger('bgpspeaker.signals.base')


class SignalBus(object):
    def __init__(self):
        self._listeners = {}

    def emit_signal(self, identifier, data):
        identifier = _to_tuple(identifier)
        LOG.debug('SIGNAL: %s emitted with data: %s ', identifier, data)
        for func, filter_func in self._listeners.get(identifier, []):
            if not filter_func or filter_func(data):
                func(identifier, data)

    def register_listener(self, identifier, func, filter_func=None):
        identifier = _to_tuple(identifier)
        substrings = (identifier[:i] for i in range(1, len(identifier) + 1))
        for partial_id in substrings:
            self._listeners.setdefault(
                partial_id,
                []
            ).append((func, filter_func))

    def unregister_all(self):
        self._listeners = {}


def _to_tuple(tuple_or_not):
    if not isinstance(tuple_or_not, tuple):
        return (tuple_or_not, )
    else:
        return tuple_or_not



# Copyright (C) 2014 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sys import intern


class CircularListType(object):
    """Instances of this class represent a specific type of list.
    Nodes are linked in a circular fashion, using attributes on the
    nodes themselves.

    Example:

      ItemList = CircularListType(next_attr='_next',
                                  prev_attr='_prev')

      l = ItemList()
      l.prepend(item)

    The created list has the following properties:

      - A node can be inserted O(1) time at the head, tail, or
        after/before another specified node.

      - A node can be removed in O(1) time from any list it may be on,
        without providing a reference to the list.

      - The current node in an iteration can be deleted safely.

     """

    class List(object):
        """An object that represents a list.

        This class is not expected to be used directly by clients. Rather, they
        would use the 'create' method of a CircularListType object to create an
        instance.
        """

        # Define the set of valid attributes so as to make the list
        # head lightweight.
        #
        # We override __getattr__ and __setattr__ so as to store the
        # the next and previous references on the list head in
        # _next_slot_ and _prev_slot_ respectively.
        __slots__ = ["list_type", "head", "_next_slot_",
                     "_prev_slot_"]

        def __init__(self, list_type):
            self.list_type = list_type

            # Save memory by using the List object itself as the head.
            self.head = self
            self.list_type.node_init(self.head)

        def __getattr__(self, name):
            if name == self.list_type.next_name:
                return self._next_slot_

            if name == self.list_type.prev_name:
                return self._prev_slot_

            raise AttributeError(name)

        def __setattr__(self, name, value):
            if name in CircularListType.List.__slots__:
                object.__setattr__(self, name, value)
                return

            if name == self.list_type.next_name:
                self._next_slot_ = value
                return

            if name == self.list_type.prev_name:
                self._prev_slot_ = value
                return

            raise AttributeError(name)

        def is_empty(self):
            return not self.list_type.node_is_on_list(self.head)

        def clear(self):
            """Remove all items from the list."""

            # Make sure that all items are unlinked.
            for node in self:
                self.remove(node)

        def is_on_list(self, node):
            return self.list_type.node_is_on_list(node)

        def append(self, node):
            self.list_type.node_insert_before(self.head, node)

        def prepend(self, node):
            self.list_type.node_insert_after(self.head, node)

        def __iter__(self):
            return self.generator()

        def remove(self, node):
            """List the given node from the list.

            Note that this does not verify that the node is on this
            list. It could even be on a different list.
            """
            self.list_type.node_unlink(node)

            self.list_type.node_del_attrs(node)

        def pop_first(self):
            """Remove the first item in the list and return it."""
            node = self.list_type.node_next(self.head)
            if node is self.head:
                return None

            self.remove(node)
            return node

        def generator(self):
            """Enables iteration over the list.

            The current item can safely be removed from the list during
            iteration.
            """
            # Keep a pointer to the next node when returning the
            # current node. This allows the caller to remove the
            # current node safely.
            node = self.list_type.node_next(self.head)
            next = self.list_type.node_next(node)
            while node is not self.head:
                yield node

                node = next
                next = self.list_type.node_next(node)

    #
    # CircularListType methods
    #

    def __init__(self, next_attr_name=None, prev_attr_name=None):
        """Initializes this list.

        next_attr_name: The name of the attribute that holds a reference
                        to the next item in the list.

        prev_attr_name: the name of the attribute that holds a reference
                        to the previous item in the list.
        """

        # Keep an interned version of the attribute names. This should
        # speed up the process of looking up the attributes.
        self.next_name = intern(next_attr_name)
        self.prev_name = intern(prev_attr_name)

    def create(self):
        return CircularListType.List(self)

    def __call__(self):
        """Make a CircularListType instance look like a class by
        creating a list object.
        """
        return self.create()

    def node_init(self, node):
        assert (not self.node_is_on_list(node))

        # Set the node to point to itself as the previous and next
        # entries.
        self.node_set_next(node, node)
        self.node_set_prev(node, node)

    def node_next(self, node):
        try:
            return getattr(node, self.next_name)
        except AttributeError:
            return None

    def node_set_next(self, node, next):
        setattr(node, self.next_name, next)

    def node_prev(self, node):
        try:
            return getattr(node, self.prev_name)
        except AttributeError:
            return None

    def node_set_prev(self, node, prev):
        setattr(node, self.prev_name, prev)

    def node_del_attrs(self, node):
        """Remove all attributes that are used for putting this node
        on this type of list.
        """
        try:
            delattr(node, self.next_name)
            delattr(node, self.prev_name)
        except AttributeError:
            pass

    def node_is_on_list(self, node):
        """Returns True if this node is on *some* list.

        A node is not on any list if it is linked to itself, or if it
        does not have the next and/prev attributes at all.
        """
        next = self.node_next(node)
        if next == node or next is None:
            assert (self.node_prev(node) is next)
            return False

        return True

    def node_insert_after(self, node, new_node):
        """Insert the new node after node."""

        assert (not self.node_is_on_list(new_node))
        assert (node is not new_node)

        next = self.node_next(node)
        assert (next is not None)
        self.node_set_next(node, new_node)
        self.node_set_prev(new_node, node)

        self.node_set_next(new_node, next)
        self.node_set_prev(next, new_node)

    def node_insert_before(self, node, new_node):
        """Insert the new node before node."""

        assert (not self.node_is_on_list(new_node))
        assert (node is not new_node)

        prev = self.node_prev(node)
        assert (prev is not None)
        self.node_set_prev(node, new_node)
        self.node_set_next(new_node, node)

        self.node_set_prev(new_node, prev)
        self.node_set_next(prev, new_node)

    def node_unlink(self, node):

        if not self.node_is_on_list(node):
            return

        prev = self.node_prev(node)
        next = self.node_next(node)

        self.node_set_next(prev, next)
        self.node_set_prev(next, prev)

        self.node_set_next(node, node)
        self.node_set_prev(node, node)



# Copyright (C) 2012 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2012 Isaku Yamahata <yamahata at private email ne jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Internal representation of datapath id is quad int
# string representation is in hex without '0x'

_DPID_LEN = 16
_DPID_FMT = '%0{0}x'.format(_DPID_LEN)
DPID_PATTERN = r'[0-9a-f]{%d}' % _DPID_LEN


def dpid_to_str(dpid):
    return _DPID_FMT % dpid


def str_to_dpid(dpid_str):
    assert len(dpid_str) == _DPID_LEN
    return int(dpid_str, 16)



# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2013 Isaku Yamahata <yamahata at private email ne jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ncclient
import ncclient.manager
import ncclient.xml_

from os_ken import exception as os_ken_exc
from os_ken.lib import of_config
from os_ken.lib.of_config import constants as ofc_consts
from os_ken.lib.of_config import classes as ofc


# TODO: When we re-organize ncclient, its NCClientError will be
#       subclass of OSKenException.
class OFConfigCapableSwitchNotFound(os_ken_exc.OSKenException,
                                    ncclient.NCClientError):
    message = 'OpenFlow Capable Switch not found'


def get_ns_tag(tag):
    if tag[0] == '{':
        return tuple(tag[1:].split('}', 1))
    return (None, tag)


class OFCapableSwitch(object):
    def __init__(self, connect_method='connect_ssh', *args, **kwargs):
        super(OFCapableSwitch, self).__init__()
        self._connect_method = connect_method
        self._connect_args = args
        self._connect_kwargs = kwargs
        self.version = None
        self.namespace = None

        connect = getattr(ncclient.manager, self._connect_method)
        self.netconf = connect(*self._connect_args, **self._connect_kwargs)

    def close_session(self):
        if self.netconf:
            self.netconf.close_session()
            self.netconf = None

    def __enter__(self):
        return self

    def __exit__(self):
        self.close_session()

    def client_capabilities(self):
        return self.netconf.client_capabilities

    def server_capabilities(self):
        return self.netconf.server_capabilities

    def _find_capable_switch(self, tree):
        capable_switch = None
        for element in tree:
            ns, tag = get_ns_tag(element.tag)
            if tag != ofc_consts.CAPABLE_SWITCH:
                continue

            # assumes that <get> returns only single capable switch
            assert capable_switch is None

            capable_switch = element
            if not self.version:
                versions = [(version, ns_) for version, ns_ in
                            of_config.OFCONFIG_YANG_NAMESPACES.items()
                            if ns == ns_]
                if versions:
                    assert len(versions) == 1
                    version = versions[0]
                    self.version, self.namespace = version

        if not capable_switch:
            raise OFConfigCapableSwitchNotFound()

        return capable_switch

    def _find_capable_switch_xml(self, tree):
        return ncclient.xml_.to_xml(self._find_capable_switch(tree))

    def raw_get(self, filter=None):
        reply = self.netconf.get(filter)
        return self._find_capable_switch_xml(reply.data_ele)

    def raw_get_config(self, source, filter=None):
        reply = self.netconf.get_config(source, filter)
        return self._find_capable_switch_xml(reply.data_ele)

    def raw_edit_config(self, target, config, default_operation=None,
                        test_option=None, error_option=None):
        self.netconf.edit_config(target, config,
                                 default_operation, test_option, error_option)

    def get(self):
        return ofc.OFCapableSwitchType.from_xml(self.raw_get())

    def get_config(self, source):
        return ofc.OFCapableSwitchType.from_xml(self.raw_get_config(source))

    def edit_config(self, target, capable_switch, default_operation=None):
        xml = ofc.NETCONF_Config(capable_switch=capable_switch).to_xml()
        self.raw_edit_config(target, xml, default_operation)

    def delete_config(self, source):
        self.netconf.delete_config(source)

    def copy_config(self, source, target):
        self.netconf.copy_config(source, target)

    def commit(self):
        self.netconf.commit()

    def discard_changes(self):
        self.netconf.discard_changes()

    # TODO: more netconf operations
    # TODO: convinience(higher level) methods



# Copyright (C) 2017 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
OpenFlow related APIs of os_ken.controller module.
"""

import netaddr

from os_ken.base import app_manager
from os_ken.lib import hub
from os_ken.lib import ip
from . import ofp_event


_TMP_ADDRESSES = {}


def register_switch_address(addr, interval=None):
    """
    Registers a new address to initiate connection to switch.

    Registers a new IP address and port pair of switch to let
    os_ken.controller.controller.OpenFlowController to try to initiate
    connection to switch.

    :param addr: A tuple of (host, port) pair of switch.
    :param interval: Interval in seconds to try to connect to switch
    """
    assert len(addr) == 2
    assert ip.valid_ipv4(addr[0]) or ip.valid_ipv6(addr[0])
    ofp_handler = app_manager.lookup_service_brick(ofp_event.NAME)
    _TMP_ADDRESSES[addr] = interval

    def _retry_loop():
        # Delays registration if ofp_handler is not started yet
        while True:
            if ofp_handler is not None and ofp_handler.controller is not None:
                for a, i in _TMP_ADDRESSES.items():
                    ofp_handler.controller.spawn_client_loop(a, i)
                    hub.sleep(1)
                break
            hub.sleep(1)

    hub.spawn(_retry_loop)


def unregister_switch_address(addr):
    """
    Unregister the given switch address.

    Unregisters the given switch address to let
    os_ken.controller.controller.OpenFlowController stop trying to initiate
    connection to switch.

    :param addr: A tuple of (host, port) pair of switch.
    """
    ofp_handler = app_manager.lookup_service_brick(ofp_event.NAME)
    # Do nothing if ofp_handler is not started yet
    if ofp_handler is None or ofp_handler.controller is None:
        return
    ofp_handler.controller.stop_client_loop(addr)



# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2013 Isaku Yamahata <yamahata at private email ne jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os_ken.base import app_manager
from os_ken.services.protocols.vrrp import event as vrrp_event


def vrrp_config(app, interface, config):
    """create an instance.
    returns EventVRRPConfigReply(instance.name, interface, config)
    on success.
    returns EventVRRPConfigReply(None, interface, config)
    on failure.
    """
    config_request = vrrp_event.EventVRRPConfigRequest(interface, config)
    config_request.sync = True
    return app.send_request(config_request)


def vrrp_shutdown(app, instance_name):
    """shutdown the instance.
    """
    shutdown_request = vrrp_event.EventVRRPShutdownRequest(instance_name)
    app.send_event(vrrp_event.VRRP_MANAGER_NAME, shutdown_request)


def vrrp_transmit(app, monitor_name, data):
    """transmit a packet from the switch.  this is internal use only.
    data is str-like, a packet to send.
    """
    transmit_request = vrrp_event.EventVRRPTransmitRequest(data)
    app.send_event(monitor_name, transmit_request)


def vrrp_list(app, instance_name=None):
    """list instances.
    returns EventVRRPListReply([VRRPInstance]).
    """
    list_request = vrrp_event.EventVRRPListRequest(instance_name)
    list_request.dst = vrrp_event.VRRP_MANAGER_NAME
    return app.send_request(list_request)


def vrrp_config_change(app, instance_name,
                       priority=None, advertisement_interval=None,
                       preempt_mode=None, accept_mode=None):
    """change configuration of an instance.
    None means no change.
    """
    config_change = vrrp_event.EventVRRPConfigChangeRequest(
        instance_name, priority, advertisement_interval,
        preempt_mode, accept_mode)
    return app.send_event(vrrp_event.VRRP_MANAGER_NAME, config_change)


app_manager.require_app('os_ken.services.protocols.vrrp.manager', api_style=True)



# Copyright (C) 2016 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import unittest

from os_ken.services.protocols.bgp.utils import validation


LOG = logging.getLogger(__name__)


class Test_Utils_Validation(unittest.TestCase):
    """
    Test case for os_ken.services.protocols.bgp.utils.validation
    """

    def test_is_valid_mac(self):
        self.assertTrue(validation.is_valid_mac('aa:bb:cc:dd:ee:ff'))

    def test_is_valid_mac_hyphenation(self):
        self.assertTrue(validation.is_valid_mac('aa-bb-cc-dd-ee-ff'))

    def test_is_valid_mac_short(self):
        self.assertEqual(False, validation.is_valid_mac('aa:bb:cc:dd:ee'))

    def test_is_valid_ip_prefix(self):
        self.assertTrue(validation.is_valid_ip_prefix(24, 32))

    def test_is_valid_ip_prefix_str(self):
        self.assertTrue(validation.is_valid_ip_prefix('24', 32))

    def test_is_valid_ip_prefix_not_digit(self):
        self.assertEqual(False, validation.is_valid_ip_prefix('foo', 32))

    def test_is_valid_ip_prefix_over(self):
        self.assertEqual(False, validation.is_valid_ip_prefix(100, 32))

    def test_is_valid_ipv4(self):
        self.assertTrue(validation.is_valid_ipv4('10.0.0.1'))

    def test_is_valid_ipv4_not_dot(self):
        self.assertEqual(False, validation.is_valid_ipv4('192:168:0:1'))

    def test_is_valid_ipv4_prefix(self):
        self.assertTrue(validation.is_valid_ipv4_prefix('10.0.0.1/24'))

    def test_is_valid_ipv4_prefix_not_str(self):
        self.assertEqual(False, validation.is_valid_ipv4_prefix(1234))

    def test_is_valid_ipv4_prefix_without_prefix(self):
        self.assertEqual(False, validation.is_valid_ipv4_prefix('10.0.0.1'))

    def test_is_valid_ipv4_prefix_invalid_addr(self):
        self.assertEqual(False,
                         validation.is_valid_ipv4_prefix('xxx.xxx.xxx.xxx/24'))

    def test_is_valid_ipv6(self):
        self.assertTrue(validation.is_valid_ipv6('fe80::0011:aabb:ccdd:eeff'))

    def test_is_valid_ipv6_not_colon(self):
        self.assertEqual(False,
                         validation.is_valid_ipv6('fe80--0011-aabb-ccdd-eeff'))

    def test_is_valid_ipv6_prefix(self):
        self.assertTrue(
            validation.is_valid_ipv6_prefix('fe80::0011:aabb:ccdd:eeff/64'))

    def test_is_valid_ipv6_prefix_not_str(self):
        self.assertEqual(False,
                         validation.is_valid_ipv6_prefix(1234))

    def test_is_valid_ipv6_prefix_without_prefix(self):
        self.assertEqual(
            False,
            validation.is_valid_ipv6_prefix('fe80::0011:aabb:ccdd:eeff'))

    def test_is_valid_ipv6_prefix_invalid_addr(self):
        self.assertEqual(False,
                         validation.is_valid_ipv6_prefix('xxxx::xxxx/64'))

    def test_is_valid_old_asn(self):
        self.assertTrue(validation.is_valid_old_asn(65000))

    def test_is_valid_old_asn_negative(self):
        self.assertEqual(False, validation.is_valid_old_asn(-1))

    def test_is_valid_old_asn_over(self):
        self.assertEqual(False, validation.is_valid_old_asn(0xffff + 1))

    def test_is_valid_asn(self):
        self.assertTrue(validation.is_valid_asn(6553800))

    def test_is_valid_asn_old(self):
        self.assertTrue(validation.is_valid_asn(65000))

    def test_is_valid_asn_negative(self):
        self.assertEqual(False, validation.is_valid_asn(-1))

    def test_is_valid_asn_over(self):
        self.assertEqual(False, validation.is_valid_asn(0xffffffff + 1))

    def test_is_valid_vpnv4_prefix(self):
        self.assertTrue(
            validation.is_valid_vpnv4_prefix('100:200:10.0.0.1/24'))

    def test_is_valid_vpnv4_prefix_not_str(self):
        self.assertEqual(False, validation.is_valid_vpnv4_prefix(1234))

    def test_is_valid_vpnv4_prefix_short_rd(self):
        self.assertEqual(False,
                         validation.is_valid_vpnv4_prefix('100:10.0.0.1/24'))

    def test_is_valid_vpnv4_prefix_invalid_rd(self):
        self.assertEqual(
            False, validation.is_valid_vpnv4_prefix('foo:bar:10.0.0.1/24'))

    def test_is_valid_vpnv6_prefix(self):
        self.assertTrue(validation.is_valid_vpnv6_prefix(
            '100:200:fe80::0011:aabb:ccdd:eeff/64'))

    def test_is_valid_vpnv6_prefix_not_str(self):
        self.assertEqual(False, validation.is_valid_vpnv6_prefix(1234))

    def test_is_valid_vpnv6_prefix_short_rd(self):
        self.assertEqual(False,
                         validation.is_valid_vpnv6_prefix('100:eeff/64'))

    def test_is_valid_vpnv6_prefix_invalid_rd(self):
        self.assertEqual(
            False, validation.is_valid_vpnv6_prefix('foo:bar:10.0.0.1/24'))

    def test_is_valid_med(self):
        self.assertTrue(validation.is_valid_med(100))

    def test_is_valid_med_not_num(self):
        self.assertEqual(False, validation.is_valid_med('foo'))

    def test_is_valid_med_negative(self):
        self.assertEqual(False, validation.is_valid_med(-1))

    def test_is_valid_med_over(self):
        self.assertEqual(False, validation.is_valid_med(0xffffffff + 1))

    def test_is_valid_mpls_label(self):
        self.assertTrue(validation.is_valid_mpls_label(100))

    def test_is_valid_mpls_label_reserved(self):
        self.assertEqual(False, validation.is_valid_mpls_label(4))

    def test_is_valid_mpls_label_not_num(self):
        self.assertEqual(False, validation.is_valid_mpls_label('foo'))

    def test_is_valid_mpls_label_negative(self):
        self.assertEqual(False, validation.is_valid_mpls_label(-1))

    def test_is_valid_mpls_label_over(self):
        self.assertEqual(False, validation.is_valid_mpls_label(0x100000 + 1))

    def test_is_valid_mpls_labels(self):
        self.assertTrue(validation.is_valid_mpls_labels([100, 200]))

    def test_is_valid_mpls_labels_not_list(self):
        self.assertEqual(False, validation.is_valid_mpls_labels(100))

    def test_is_valid_mpls_labels_with_invalid_label(self):
        self.assertEqual(False, validation.is_valid_mpls_labels(['foo', 200]))

    def test_is_valid_route_dist(self):
        self.assertTrue(validation.is_valid_route_dist('65000:222'))

    def test_is_valid_route_dist_ipv4_based(self):
        self.assertTrue(validation.is_valid_route_dist('10.0.0.1:333'))

    def test_is_valid_route_not_str(self):
        self.assertEqual(False, validation.is_valid_route_dist(65000))

    def test_is_valid_route_dist_short(self):
        self.assertEqual(False, validation.is_valid_route_dist('65000'))

    def test_is_valid_route_dist_invalid_ipv4_addr(self):
        self.assertEqual(False,
                         validation.is_valid_route_dist('xxx.xxx.xxx.xxx:333'))

    def test_is_valid_esi(self):
        self.assertTrue(validation.is_valid_esi(100))

    def test_is_valid_esi_not_int(self):
        self.assertEqual(False, validation.is_valid_esi('foo'))

    def test_is_valid_ethernet_tag_id(self):
        self.assertTrue(validation.is_valid_ethernet_tag_id(100))

    def test_is_valid_ethernet_tag_id_not_int(self):
        self.assertEqual(False, validation.is_valid_ethernet_tag_id('foo'))

    def test_is_valid_ethernet_tag_id_negative(self):
        self.assertEqual(False, validation.is_valid_ethernet_tag_id(-1))

    def test_is_valid_ethernet_tag_id_over(self):
        self.assertEqual(False,
                         validation.is_valid_ethernet_tag_id(0xffffffff + 1))

    def test_is_valid_vni(self):
        self.assertTrue(validation.is_valid_vni(100))

    def test_is_valid_vni_not_int(self):
        self.assertEqual(False, validation.is_valid_vni('foo'))

    def test_is_valid_vni_negative(self):
        self.assertEqual(False, validation.is_valid_vni(-1))

    def test_is_valid_vni_over(self):
        self.assertEqual(False, validation.is_valid_vni(0xffffff + 1))



# Copyright (C) 2014 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
 Runtime configuration that applies to all bgp sessions, i.e. global settings.
"""
import logging
import numbers

from os_ken.lib import ip

from os_ken.services.protocols.bgp.utils.validation import is_valid_ipv4
from os_ken.services.protocols.bgp.utils.validation import is_valid_asn

from os_ken.services.protocols.bgp import rtconf
from os_ken.services.protocols.bgp.rtconf.base import BaseConf
from os_ken.services.protocols.bgp.rtconf.base import BaseConfListener
from os_ken.services.protocols.bgp.rtconf.base import compute_optional_conf
from os_ken.services.protocols.bgp.rtconf.base import ConfigTypeError
from os_ken.services.protocols.bgp.rtconf.base import ConfigValueError
from os_ken.services.protocols.bgp.rtconf.base import MissingRequiredConf
from os_ken.services.protocols.bgp.rtconf.base import validate

LOG = logging.getLogger('bgpspeaker.rtconf.common')


# Global configuration settings.
LOCAL_AS = 'local_as'
ROUTER_ID = 'router_id'
CLUSTER_ID = 'cluster_id'
LABEL_RANGE = 'label_range'
LABEL_RANGE_MAX = 'max'
LABEL_RANGE_MIN = 'min'
LOCAL_PREF = 'local_pref'

# Similar to Cisco command 'allowas-in'. Allows the local ASN in the path.
# Facilitates auto rd, auto rt import/export
# ("rd auto/route-target both auto") and simplified spine/leaf architectures,
# sharing an ASN between e.g. leafs.
ALLOW_LOCAL_AS_IN_COUNT = 'allow_local_as_in_count'

# Configuration that can be set at global level as well as per context
# (session/vrf) level
# Nested configuration override global or higher level configuration as they
# are more granular.
# TODO(apgw-dev) Nested configuration overriding higher level configuration is
# currently low priority

# Similar to Cisco command 'bgp refresh stalepath-time'. To cause the router to
# remove stale routes from the BGP table even if the router does not receive a
# Route-Refresh EOR message The bgp refresh stalepath-time command is not
# needed under normal circumstances.
# TODO(PH): Support this feature (currently low priority)
REFRESH_STALEPATH_TIME = 'refresh_stalepath_time'

# Similar to Cisco command 'bgp refresh max-eor-time'. The bgp refresh max-eor-
# time command is not needed under normal  circumstances. You might configure
# the bgp refresh max-eor-time command in the event of continuous route
# flapping, when the router is unable to generate a Route- Refresh EOR message,
# in which case a Route-Refresh EOR is generated after the timer expires.
# TODO(PH): Support this feature (currently low priority)
REFRESH_MAX_EOR_TIME = 'refresh_max_eor_time'

BGP_CONN_RETRY_TIME = 'bgp_conn_retry_time'
BGP_SERVER_HOSTS = 'bgp_server_hosts'
BGP_SERVER_PORT = 'bgp_server_port'
TCP_CONN_TIMEOUT = 'tcp_conn_timeout'
MAX_PATH_EXT_RTFILTER_ALL = 'maximum_paths_external_rtfilter_all'


# Valid default values of some settings.
DEFAULT_LABEL_RANGE = (100, 100000)
DEFAULT_REFRESH_STALEPATH_TIME = 0
DEFAULT_REFRESH_MAX_EOR_TIME = 0
DEFAULT_BGP_SERVER_HOSTS = ('0.0.0.0', '::')
DEFAULT_BGP_SERVER_PORT = 179
DEFAULT_TCP_CONN_TIMEOUT = 30
DEFAULT_BGP_CONN_RETRY_TIME = 30
DEFAULT_MED = 0
DEFAULT_MAX_PATH_EXT_RTFILTER_ALL = True
DEFAULT_LOCAL_PREF = 100


@validate(name=ALLOW_LOCAL_AS_IN_COUNT)
def validate_allow_local_as_in_count(count):
    if not isinstance(count, numbers.Integral):
        raise ConfigTypeError(desc=('Configuration value for %s has to be '
                                    'integral type' % ALLOW_LOCAL_AS_IN_COUNT))
    if count < 0:
        raise ConfigValueError(desc='Invalid local AS count %s' % count)

    return count


@validate(name=LOCAL_AS)
def validate_local_as(asn):
    if asn is None:
        raise MissingRequiredConf(conf_name=LOCAL_AS)

    if not is_valid_asn(asn):
        raise ConfigValueError(desc='Invalid local_as configuration value: %s'
                               % asn)
    return asn


@validate(name=ROUTER_ID)
def validate_router_id(router_id):
    if not router_id:
        raise MissingRequiredConf(conf_name=ROUTER_ID)

    if not isinstance(router_id, str):
        raise ConfigTypeError(conf_name=ROUTER_ID)
    if not is_valid_ipv4(router_id):
        raise ConfigValueError(desc='Invalid router id %s' % router_id)

    return router_id


@validate(name=CLUSTER_ID)
def validate_router_id(cluster_id):
    if not isinstance(cluster_id, str):
        raise ConfigTypeError(conf_name=CLUSTER_ID)
    if not is_valid_ipv4(cluster_id):
        raise ConfigValueError(desc='Invalid cluster id %s' % cluster_id)

    return cluster_id


@validate(name=REFRESH_STALEPATH_TIME)
def validate_refresh_stalepath_time(rst):
    if not isinstance(rst, numbers.Integral):
        raise ConfigTypeError(desc=('Configuration value for %s has to be '
                                    'integral type' % REFRESH_STALEPATH_TIME))
    if rst < 0:
        raise ConfigValueError(desc='Invalid refresh stalepath time %s' % rst)

    return rst


@validate(name=REFRESH_MAX_EOR_TIME)
def validate_refresh_max_eor_time(rmet):
    if not isinstance(rmet, numbers.Integral):
        raise ConfigTypeError(desc=('Configuration value for %s has to be of '
                                    'integral type ' % REFRESH_MAX_EOR_TIME))
    if rmet < 0:
        raise ConfigValueError(desc='Invalid refresh stalepath time %s' % rmet)

    return rmet


@validate(name=LABEL_RANGE)
def validate_label_range(label_range):
    min_label, max_label = label_range
    if (not min_label or
            not max_label or
            not isinstance(min_label, numbers.Integral) or
            not isinstance(max_label, numbers.Integral) or min_label < 17 or
            min_label >= max_label):
        raise ConfigValueError(desc=('Invalid label_range configuration value:'
                                     ' (%s).' % label_range))

    return label_range


@validate(name=BGP_SERVER_HOSTS)
def validate_bgp_server_hosts(hosts):
    for host in hosts:
        if not ip.valid_ipv4(host) and not ip.valid_ipv6(host):
            raise ConfigTypeError(desc=('Invalid bgp sever hosts '
                                        'configuration value %s' % hosts))

    return hosts


@validate(name=BGP_SERVER_PORT)
def validate_bgp_server_port(server_port):
    if not isinstance(server_port, numbers.Integral):
        raise ConfigTypeError(desc=('Invalid bgp sever port configuration '
                                    'value %s' % server_port))
    if server_port < 0 or server_port > 65535:
        raise ConfigValueError(desc='Invalid server port %s' % server_port)

    return server_port


@validate(name=TCP_CONN_TIMEOUT)
def validate_tcp_conn_timeout(tcp_conn_timeout):
    # TODO(apgw-dev) made-up some valid values for this settings, check if we
    # have a standard value in any routers
    if not isinstance(tcp_conn_timeout, numbers.Integral):
        raise ConfigTypeError(desc=('Invalid tcp connection timeout '
                                    'configuration value %s' %
                                    tcp_conn_timeout))

    if tcp_conn_timeout < 10:
        raise ConfigValueError(desc=('Invalid tcp connection timeout'
                                     ' configuration value %s' %
                                     tcp_conn_timeout))

    return tcp_conn_timeout


@validate(name=BGP_CONN_RETRY_TIME)
def validate_bgp_conn_retry_time(bgp_conn_retry_time):
    if not isinstance(bgp_conn_retry_time, numbers.Integral):
        raise ConfigTypeError(desc=('Invalid bgp conn. retry time '
                                    'configuration value %s' %
                                    bgp_conn_retry_time))

    if bgp_conn_retry_time < 10:
        raise ConfigValueError(desc=('Invalid bgp connection retry time'
                                     ' configuration value %s' %
                                     bgp_conn_retry_time))
    return bgp_conn_retry_time


@validate(name=MAX_PATH_EXT_RTFILTER_ALL)
def validate_max_path_ext_rtfilter_all(max_path_ext_rtfilter_all):
    if not isinstance(max_path_ext_rtfilter_all, bool):
        raise ConfigTypeError(desc=('Invalid max_path_ext_rtfilter_all'
                                    ' configuration value %s' %
                                    max_path_ext_rtfilter_all))
    return max_path_ext_rtfilter_all


@validate(name=LOCAL_PREF)
def validate_local_pref(local_pref):
    if not isinstance(local_pref, numbers.Integral):
        raise ConfigTypeError(desc=('Invalid local_pref'
                                    ' configuration value %s' %
                                    local_pref))
    return local_pref


class CommonConf(BaseConf):
    """Encapsulates configurations applicable to all peer sessions.

    Currently if any of these configurations change, it is assumed that current
    active peer session will be bought down and restarted.
    """
    CONF_CHANGED_EVT = 1

    VALID_EVT = frozenset([CONF_CHANGED_EVT])

    REQUIRED_SETTINGS = frozenset([ROUTER_ID, LOCAL_AS])

    OPTIONAL_SETTINGS = frozenset([REFRESH_STALEPATH_TIME,
                                   REFRESH_MAX_EOR_TIME, LABEL_RANGE,
                                   BGP_SERVER_HOSTS, BGP_SERVER_PORT,
                                   TCP_CONN_TIMEOUT,
                                   BGP_CONN_RETRY_TIME,
                                   MAX_PATH_EXT_RTFILTER_ALL,
                                   ALLOW_LOCAL_AS_IN_COUNT,
                                   CLUSTER_ID,
                                   LOCAL_PREF])

    def __init__(self, **kwargs):
        super(CommonConf, self).__init__(**kwargs)

    def _init_opt_settings(self, **kwargs):
        super(CommonConf, self)._init_opt_settings(**kwargs)
        self._settings[ALLOW_LOCAL_AS_IN_COUNT] = compute_optional_conf(
            ALLOW_LOCAL_AS_IN_COUNT, 0, **kwargs)
        self._settings[LABEL_RANGE] = compute_optional_conf(
            LABEL_RANGE, DEFAULT_LABEL_RANGE, **kwargs)
        self._settings[REFRESH_STALEPATH_TIME] = compute_optional_conf(
            REFRESH_STALEPATH_TIME, DEFAULT_REFRESH_STALEPATH_TIME, **kwargs)
        self._settings[REFRESH_MAX_EOR_TIME] = compute_optional_conf(
            REFRESH_MAX_EOR_TIME, DEFAULT_REFRESH_MAX_EOR_TIME, **kwargs)
        self._settings[BGP_SERVER_HOSTS] = compute_optional_conf(
            BGP_SERVER_HOSTS, DEFAULT_BGP_SERVER_HOSTS, **kwargs)
        self._settings[BGP_SERVER_PORT] = compute_optional_conf(
            BGP_SERVER_PORT, DEFAULT_BGP_SERVER_PORT, **kwargs)
        self._settings[TCP_CONN_TIMEOUT] = compute_optional_conf(
            TCP_CONN_TIMEOUT, DEFAULT_TCP_CONN_TIMEOUT, **kwargs)
        self._settings[BGP_CONN_RETRY_TIME] = compute_optional_conf(
            BGP_CONN_RETRY_TIME, DEFAULT_BGP_CONN_RETRY_TIME, **kwargs)
        self._settings[MAX_PATH_EXT_RTFILTER_ALL] = compute_optional_conf(
            MAX_PATH_EXT_RTFILTER_ALL, DEFAULT_MAX_PATH_EXT_RTFILTER_ALL,
            **kwargs)
        self._settings[CLUSTER_ID] = compute_optional_conf(
            CLUSTER_ID, kwargs[ROUTER_ID], **kwargs)
        self._settings[LOCAL_PREF] = compute_optional_conf(
            LOCAL_PREF, DEFAULT_LOCAL_PREF, **kwargs)

    # =========================================================================
    # Required attributes
    # =========================================================================

    @property
    def local_as(self):
        return self._settings[LOCAL_AS]

    @property
    def router_id(self):
        return self._settings[ROUTER_ID]

    # =========================================================================
    # Optional attributes with valid defaults.
    # =========================================================================
    @property
    def cluster_id(self):
        return self._settings[CLUSTER_ID]

    @property
    def allow_local_as_in_count(self):
        return self._settings[ALLOW_LOCAL_AS_IN_COUNT]

    @property
    def bgp_conn_retry_time(self):
        return self._settings[BGP_CONN_RETRY_TIME]

    @property
    def tcp_conn_timeout(self):
        return self._settings[TCP_CONN_TIMEOUT]

    @property
    def refresh_stalepath_time(self):
        return self._settings[REFRESH_STALEPATH_TIME]

    @property
    def refresh_max_eor_time(self):
        return self._settings[REFRESH_MAX_EOR_TIME]

    @property
    def label_range(self):
        return self._settings[LABEL_RANGE]

    @property
    def bgp_server_hosts(self):
        return self._settings[BGP_SERVER_HOSTS]

    @property
    def bgp_server_port(self):
        return self._settings[BGP_SERVER_PORT]

    @property
    def max_path_ext_rtfilter_all(self):
        return self._settings[MAX_PATH_EXT_RTFILTER_ALL]

    @property
    def local_pref(self):
        return self._settings[LOCAL_PREF]

    @classmethod
    def get_opt_settings(self):
        self_confs = super(CommonConf, self).get_opt_settings()
        self_confs.update(CommonConf.OPTIONAL_SETTINGS)
        return self_confs

    @classmethod
    def get_req_settings(self):
        self_confs = super(CommonConf, self).get_req_settings()
        self_confs.update(CommonConf.REQUIRED_SETTINGS)
        return self_confs

    @classmethod
    def get_valid_evts(self):
        self_valid_evts = super(CommonConf, self).get_valid_evts()
        self_valid_evts.update(CommonConf.VALID_EVT)
        return self_valid_evts

    def update(self, **kwargs):
        """Updates global configuration settings with given values.

        First checks if given configuration values differ from current values.
        If any of the configuration values changed, generates a change event.
        Currently we generate change event for any configuration change.
        Note: This method is idempotent.
        """
        # Update inherited configurations
        super(CommonConf, self).update(**kwargs)
        conf_changed = False

        # Validate given configurations and check if value changed
        for conf_name, conf_value in kwargs.items():
            rtconf.base.get_validator(conf_name)(conf_value)
            item1 = self._settings.get(conf_name, None)
            item2 = kwargs.get(conf_name, None)

            if item1 != item2:
                conf_changed = True

        # If any configuration changed, we update configuration value and
        # notify listeners
        if conf_changed:
            for conf_name, conf_value in kwargs.items():
                # Since all new values are already validated, we can use them
                self._settings[conf_name] = conf_value

            self._notify_listeners(CommonConf.CONF_CHANGED_EVT, self)


class CommonConfListener(BaseConfListener):
    """Base listener for various changes to common configurations."""

    def __init__(self, global_conf):
        super(CommonConfListener, self).__init__(global_conf)
        global_conf.add_listener(CommonConf.CONF_CHANGED_EVT,
                                 self.on_update_common_conf)

    def on_update_common_conf(self, evt):
        raise NotImplementedError('This method should be overridden.')



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import sys
from setuptools.command import easy_install
from os_ken import version


# Global variables in this module doesn't work as we expect
# because, during the setup procedure, this module seems to be
# copied (as a file) and can be loaded multiple times.
# We save them into __main__ module instead.
def _main_module():
    return sys.modules['__main__']


def save_orig():
    """Save original easy_install.get_script_args.
    This is necessary because pbr's setup_hook is sometimes called
    before ours."""
    _main_module()._orig_get_script_args = easy_install.get_script_args


def setup_hook(config):
    """Filter config parsed from a setup.cfg to inject our defaults."""
    metadata = config['metadata']
    if sys.platform == 'win32':
        requires = metadata.get('requires_dist', '').split('\n')
        metadata['requires_dist'] = "\n".join(requires)
    config['metadata'] = metadata

    metadata['version'] = str(version)

    # pbr's setup_hook replaces easy_install.get_script_args with
    # their own version, override_get_script_args, prefering simpler
    # scripts which are not aware of multi-version.
    # prevent that by doing the opposite.  it's a horrible hack
    # but we are in patching wars already...
    from pbr import packaging

    def my_get_script_args(*args, **kwargs):
        return _main_module()._orig_get_script_args(*args, **kwargs)

    packaging.override_get_script_args = my_get_script_args
    easy_install.get_script_args = my_get_script_args

    # another hack to allow setup from tarball.
    orig_get_version = packaging.get_version

    def my_get_version(package_name, pre_version=None):
        if package_name == 'os_ken':
            return str(version)
        return orig_get_version(package_name, pre_version)

    packaging.get_version = my_get_version



# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from . import packet_base
from os_ken.lib import addrconv

# Slow Protocol Multicast destination
SLOW_PROTOCOL_MULTICAST = '01:80:c2:00:00:02'

# Slow Protocol SubType
SLOW_SUBTYPE_LACP = 0x01
SLOW_SUBTYPE_MARKER = 0x02
SLOW_SUBTYPE_OAM = 0x03
SLOW_SUBTYPE_OSSP = 0x0a


class slow(packet_base.PacketBase):
    """Slow Protocol header decoder class.
    This class has only the parser method.

    http://standards.ieee.org/getieee802/download/802.3-2012_section5.pdf

    Slow Protocols Subtypes

    +---------------+--------------------------------------------------+
    | Subtype Value | Protocol Name                                    |
    +===============+==================================================+
    | 0             | Unused - Illegal Value                           |
    +---------------+--------------------------------------------------+
    | 1             | Link Aggregation Control Protocol(LACP)          |
    +---------------+--------------------------------------------------+
    | 2             | Link Aggregation - Marker Protocol               |
    +---------------+--------------------------------------------------+
    | 3             | Operations, Administration, and Maintenance(OAM) |
    +---------------+--------------------------------------------------+
    | 4 - 9         | Reserved for future use                          |
    +---------------+--------------------------------------------------+
    | 10            | Organization Specific Slow Protocol(OSSP)        |
    +---------------+--------------------------------------------------+
    | 11 - 255      | Unused - Illegal values                          |
    +---------------+--------------------------------------------------+
    """
    _PACK_STR = '!B'

    @classmethod
    def parser(cls, buf):
        (subtype, ) = struct.unpack_from(cls._PACK_STR, buf)
        switch = {
            SLOW_SUBTYPE_LACP: lacp,
            # TODO: make parsers of other subtypes.
            SLOW_SUBTYPE_MARKER: None,
            SLOW_SUBTYPE_OAM: None,
            SLOW_SUBTYPE_OSSP: None,
        }
        cls_ = switch.get(subtype)
        if cls_:
            return cls_.parser(buf)
        else:
            return None, None, buf


class lacp(packet_base.PacketBase):
    """Link Aggregation Control Protocol(LACP, IEEE 802.1AX)
    header encoder/decoder class.

    http://standards.ieee.org/getieee802/download/802.1AX-2008.pdf

    LACPDU format

    +------------------------------------------------+--------+
    | LACPDU structure                               | Octets |
    +================================================+========+
    | Subtype = LACP                                 | 1      |
    +------------------------------------------------+--------+
    | Version Number                                 | 1      |
    +------------+-----------------------------------+--------+
    | TLV        | TLV_type = Actor Information      | 1      |
    | Actor      |                                   |        |
    +------------+-----------------------------------+--------+
    |            | Actor_Information_Length = 20     | 1      |
    +------------+-----------------------------------+--------+
    |            | Actor_System_Priority             | 2      |
    +------------+-----------------------------------+--------+
    |            | Actor_System                      | 6      |
    +------------+-----------------------------------+--------+
    |            | Actor_Key                         | 2      |
    +------------+-----------------------------------+--------+
    |            | Actor_Port_Priority               | 2      |
    +------------+-----------------------------------+--------+
    |            | Actor_Port                        | 2      |
    +------------+-----------------------------------+--------+
    |            | Actor_State                       | 1      |
    +------------+-----------------------------------+--------+
    |            | Reserved                          | 3      |
    +------------+-----------------------------------+--------+
    | TLV        | TLV_type = Partner Information    | 1      |
    | Partner    |                                   |        |
    +------------+-----------------------------------+--------+
    |            | Partner_Information_Length = 20   | 1      |
    +------------+-----------------------------------+--------+
    |            | Partner_System_Priority           | 2      |
    +------------+-----------------------------------+--------+
    |            | Partner_System                    | 6      |
    +------------+-----------------------------------+--------+
    |            | Partner_Key                       | 2      |
    +------------+-----------------------------------+--------+
    |            | Partner_Port_Priority             | 2      |
    +------------+-----------------------------------+--------+
    |            | Partner_Port                      | 2      |
    +------------+-----------------------------------+--------+
    |            | Partner_State                     | 1      |
    +------------+-----------------------------------+--------+
    |            | Reserved                          | 3      |
    +------------+-----------------------------------+--------+
    | TLV        | TLV_type = Collector Information  | 1      |
    | Collector  |                                   |        |
    +------------+-----------------------------------+--------+
    |            | Collector_Information_Length = 16 | 1      |
    +------------+-----------------------------------+--------+
    |            | Collector_Max_Delay               | 2      |
    +------------+-----------------------------------+--------+
    |            | Reserved                          | 12     |
    +------------+-----------------------------------+--------+
    | TLV        | TLV_type = Terminator             | 1      |
    | Terminator |                                   |        |
    +------------+-----------------------------------+--------+
    |            | Terminator_Length = 0             | 1      |
    +------------+-----------------------------------+--------+
    |            | Reserved                          | 50     |
    +------------+-----------------------------------+--------+


    Terminator information uses a length value of 0 (0x00).

    NOTE--The use of a Terminator_Length of 0 is intentional.
          In TLV encoding schemes it is common practice
          for the terminator encoding to be 0 both
          for the type and the length.

    Actor_State and Partner_State encoded as individual bits within
    a single octet as follows:

    +------+------+------+------+------+------+------+------+
    | 7    | 6    | 5    | 4    | 3    | 2    | 1    | 0    |
    +======+======+======+======+======+======+======+======+
    | EXPR | DFLT | DIST | CLCT | SYNC | AGGR | TMO  | ACT  |
    +------+------+------+------+------+------+------+------+

    ACT
        bit 0.
        about the activity control value with regard to this link.
    TMO
        bit 1.
        about the timeout control value with regard to this link.
    AGGR
        bit 2.
        about how the system regards this link from the point of view
        of the aggregation.
    SYNC
        bit 3.
        about how the system regards this link from the point of view
        of the synchronization.
    CLCT
        bit 4.
        about collecting of incoming frames.
    DIST
        bit 5.
        about distributing of outgoing frames.
    DFLT
        bit 6.
        about the opposite system information which the system use.
    EXPR
        bit 7.
        about the expire state of the system.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte
    order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    =============================== ====================================
    Attribute                       Description
    =============================== ====================================
    version                         LACP version. This parameter must be
                                    set to LACP_VERSION_NUMBER(i.e. 1).

    actor_system_priority           The priority assigned to this
                                    System.

    actor_system                    The Actor's System ID, encoded as
                                    a MAC address.

    actor_key                       The operational Key value assigned
                                    to the port by the Actor.

    actor_port_priority             The priority assigned to this port.

    actor_port                      The port number assigned to the
                                    port by the Actor.

    actor_state_activity            .. _lacp_activity:

                                    about the activity control value
                                    with regard to this link.

                                    LACP_STATE_ACTIVE(1)

                                    LACP_STATE_PASSIVE(0)

    actor_state_timeout             .. _lacp_timeout:

                                    about the timeout control value
                                    with regard to this link.

                                    LACP_STATE_SHORT_TIMEOUT(1)

                                    LACP_STATE_LONG_TIMEOUT(0)

    actor_state_aggregation         .. _lacp_aggregation:

                                    about how the system regards this
                                    link from the point of view of the
                                    aggregation.

                                    LACP_STATE_AGGREGATEABLE(1)

                                    LACP_STATE_INDIVIDUAL(0)

    actor_state_synchronization     .. _lacp_synchronization:

                                    about how the system regards this
                                    link from the point of view of the
                                    synchronization.

                                    LACP_STATE_IN_SYNC(1)

                                    LACP_STATE_OUT_OF_SYNC(0)

    actor_state_collecting          .. _lacp_collecting:

                                    about collecting of incoming frames.

                                    LACP_STATE_COLLECTING_ENABLED(1)

                                    LACP_STATE_COLLECTING_DISABLED(0)

    actor_state_distributing        .. _lacp_distributing:

                                    about distributing of outgoing frames.

                                    LACP_STATE_DISTRIBUTING_ENABLED(1)

                                    LACP_STATE_DISTRIBUTING_DISABLED(0)

    actor_state_defaulted           .. _lacp_defaulted:

                                    about the Partner information
                                    which the the Actor use.

                                    LACP_STATE_DEFAULTED_PARTNER(1)

                                    LACP_STATE_OPERATIONAL_PARTNER(0)

    actor_state_expired             .. _lacp_expired:

                                    about the state of the Actor.

                                    LACP_STATE_EXPIRED(1)

                                    LACP_STATE_NOT_EXPIRED(0)

    partner_system_priority         The priority assigned to the
                                    Partner System.

    partner_system                  The Partner's System ID, encoded
                                    as a MAC address.

    partner_key                     The operational Key value assigned
                                    to the port by the Partner.

    partner_port_priority           The priority assigned to this port
                                    by the Partner.

    partner_port                    The port number assigned to the
                                    port by the Partner.

    partner_state_activity          See :ref:`actor_state_activity\
                                    <lacp_activity>`.

    partner_state_timeout           See :ref:`actor_state_timeout\
                                    <lacp_timeout>`.

    partner_state_aggregation       See :ref:`actor_state_aggregation\
                                    <lacp_aggregation>`.

    partner_state_synchronization   See
                                    :ref:`actor_state_synchronization\
                                    <lacp_synchronization>`.

    partner_state_collecting        See :ref:`actor_state_collecting\
                                    <lacp_collecting>`.

    partner_state_distributing      See :ref:`actor_state_distributing\
                                    <lacp_distributing>`.

    partner_state_defaulted         See :ref:`actor_state_defaulted\
                                    <lacp_defaulted>`.

    partner_state_expired           See :ref:`actor_state_expired\
                                    <lacp_expired>`.

    collector_max_delay             the maximum time that the Frame
                                    Collector may delay.
    =============================== ====================================

    """
    LACP_VERSION_NUMBER = 1

    # LACP TLV type
    LACP_TLV_TYPE_ACTOR = 1
    LACP_TLV_TYPE_PARTNER = 2
    LACP_TLV_TYPE_COLLECTOR = 3
    LACP_TLV_TYPE_TERMINATOR = 0

    # LACP state(LACP_Activity)
    LACP_STATE_ACTIVE = 1
    LACP_STATE_PASSIVE = 0
    # LACP state(LACP_Timeout)
    LACP_STATE_SHORT_TIMEOUT = 1
    LACP_STATE_LONG_TIMEOUT = 0
    # LACP state(Aggregation)
    LACP_STATE_AGGREGATEABLE = 1
    LACP_STATE_INDIVIDUAL = 0
    # LACP state(Synchronization)
    LACP_STATE_IN_SYNC = 1
    LACP_STATE_OUT_OF_SYNC = 0
    # LACP state(Collecting)
    LACP_STATE_COLLECTING_ENABLED = 1
    LACP_STATE_COLELCTING_DISABLED = 0
    # LACP state(Distributing)
    LACP_STATE_DISTRIBUTING_ENABLED = 1
    LACP_STATE_DISTRIBUTING_DISABLED = 0
    # LACP state(Defaulted)
    LACP_STATE_DEFAULED_PARTNER = 1
    LACP_STATE_OPERATIONAL_PARTNER = 0
    # LACP state(Expired)
    LACP_STATE_EXPIRED = 1
    LACP_STATE_NOT_EXPIRED = 0

    # The number of seconds between periodic transmissions using
    # Short Timeouts.
    FAST_PERIODIC_TIME = 1
    # The number of seconds between periodic transmissions using
    # Long Timeouts.
    SLOW_PERIODIC_TIME = 30
    # The number of seconds before invalidating received LACPDU
    # information when using Short Timeouts(3 x Fast_Periodic_Time).
    SHORT_TIMEOUT_TIME = 3 * FAST_PERIODIC_TIME
    # The number of seconds before invalidating received LACPDU
    # information when using Long Timeouts (3 x Slow_Periodic_Time).
    LONG_TIMEOUT_TIME = 3 * SLOW_PERIODIC_TIME

    _HLEN_PACK_STR = '!BB'
    _HLEN_PACK_LEN = struct.calcsize(_HLEN_PACK_STR)
    _ACTPRT_INFO_PACK_STR = '!BBH6sHHHB3x'
    _ACTPRT_INFO_PACK_LEN = struct.calcsize(_ACTPRT_INFO_PACK_STR)
    _COL_INFO_PACK_STR = '!BBH12x'
    _COL_INFO_PACK_LEN = struct.calcsize(_COL_INFO_PACK_STR)
    _TRM_PACK_STR = '!BB50x'
    _TRM_PACK_LEN = struct.calcsize(_TRM_PACK_STR)
    _ALL_PACK_LEN = _HLEN_PACK_LEN + _ACTPRT_INFO_PACK_LEN * 2 + \
        _COL_INFO_PACK_LEN + _TRM_PACK_LEN

    _MIN_LEN = _ALL_PACK_LEN

    _TYPE = {
        'ascii': [
            'actor_system', 'partner_system'
        ]
    }

    def __init__(self, version=LACP_VERSION_NUMBER,
                 actor_system_priority=0,
                 actor_system='00:00:00:00:00:00',
                 actor_key=0, actor_port_priority=0, actor_port=0,
                 actor_state_activity=0, actor_state_timeout=0,
                 actor_state_aggregation=0,
                 actor_state_synchronization=0,
                 actor_state_collecting=0, actor_state_distributing=0,
                 actor_state_defaulted=0, actor_state_expired=0,
                 partner_system_priority=0,
                 partner_system='00:00:00:00:00:00',
                 partner_key=0, partner_port_priority=0, partner_port=0,
                 partner_state_activity=0, partner_state_timeout=0,
                 partner_state_aggregation=0,
                 partner_state_synchronization=0,
                 partner_state_collecting=0,
                 partner_state_distributing=0,
                 partner_state_defaulted=0, partner_state_expired=0,
                 collector_max_delay=0):
        super(lacp, self).__init__()
        # parameter check
        assert (1 == actor_state_activity | 1)
        assert (1 == actor_state_timeout | 1)
        assert (1 == actor_state_aggregation | 1)
        assert (1 == actor_state_synchronization | 1)
        assert (1 == actor_state_collecting | 1)
        assert (1 == actor_state_distributing | 1)
        assert (1 == actor_state_defaulted | 1)
        assert (1 == actor_state_expired | 1)
        assert (1 == partner_state_activity | 1)
        assert (1 == partner_state_timeout | 1)
        assert (1 == partner_state_aggregation | 1)
        assert (1 == partner_state_synchronization | 1)
        assert (1 == partner_state_collecting | 1)
        assert (1 == partner_state_distributing | 1)
        assert (1 == partner_state_defaulted | 1)
        assert (1 == partner_state_expired | 1)
        # ------------------------------
        # Header
        # ------------------------------
        self._subtype = SLOW_SUBTYPE_LACP
        self.version = version
        # ------------------------------
        # Actor Information
        # ------------------------------
        self._actor_tag = self.LACP_TLV_TYPE_ACTOR
        self._actor_length = self._ACTPRT_INFO_PACK_LEN
        self.actor_system_priority = actor_system_priority
        self.actor_system = actor_system
        self.actor_key = actor_key
        self.actor_port_priority = actor_port_priority
        self.actor_port = actor_port
        self.actor_state_activity = actor_state_activity
        self.actor_state_timeout = actor_state_timeout
        self.actor_state_aggregation = actor_state_aggregation
        self.actor_state_synchronization = actor_state_synchronization
        self.actor_state_collecting = actor_state_collecting
        self.actor_state_distributing = actor_state_distributing
        self.actor_state_defaulted = actor_state_defaulted
        self.actor_state_expired = actor_state_expired
        self._actor_state = (
            (self.actor_state_activity << 0) |
            (self.actor_state_timeout << 1) |
            (self.actor_state_aggregation << 2) |
            (self.actor_state_synchronization << 3) |
            (self.actor_state_collecting << 4) |
            (self.actor_state_distributing << 5) |
            (self.actor_state_defaulted << 6) |
            (self.actor_state_expired << 7))
        # ------------------------------
        # Partner Information
        # ------------------------------
        self._partner_tag = self.LACP_TLV_TYPE_PARTNER
        self._partner_length = self._ACTPRT_INFO_PACK_LEN
        self.partner_system_priority = partner_system_priority
        self.partner_system = partner_system
        self.partner_key = partner_key
        self.partner_port_priority = partner_port_priority
        self.partner_port = partner_port
        self.partner_state_activity = partner_state_activity
        self.partner_state_timeout = partner_state_timeout
        self.partner_state_aggregation = partner_state_aggregation
        self.partner_state_synchronization = \
            partner_state_synchronization
        self.partner_state_collecting = partner_state_collecting
        self.partner_state_distributing = partner_state_distributing
        self.partner_state_defaulted = partner_state_defaulted
        self.partner_state_expired = partner_state_expired
        self._partner_state = (
            (self.partner_state_activity << 0) |
            (self.partner_state_timeout << 1) |
            (self.partner_state_aggregation << 2) |
            (self.partner_state_synchronization << 3) |
            (self.partner_state_collecting << 4) |
            (self.partner_state_distributing << 5) |
            (self.partner_state_defaulted << 6) |
            (self.partner_state_expired << 7))
        # ------------------------------
        # Collector Information
        # ------------------------------
        self._collector_tag = self.LACP_TLV_TYPE_COLLECTOR
        self._collector_length = self._COL_INFO_PACK_LEN
        self.collector_max_delay = collector_max_delay
        # ------------------------------
        # Terminator
        # ------------------------------
        self._terminator_tag = self.LACP_TLV_TYPE_TERMINATOR
        self._terminator_length = 0

    @classmethod
    def parser(cls, buf):
        assert cls._ALL_PACK_LEN == len(buf)
        offset = 0
        # ------------------------------
        # Header
        # ------------------------------
        (subtype, version
         ) = struct.unpack_from(cls._HLEN_PACK_STR, buf, offset)
        assert SLOW_SUBTYPE_LACP == subtype
        assert cls.LACP_VERSION_NUMBER == version
        offset += cls._HLEN_PACK_LEN
        # ------------------------------
        # Actor Information
        # ------------------------------
        (actor_tag, actor_length, actor_system_priority, actor_system,
         actor_key, actor_port_priority, actor_port, actor_state
         ) = struct.unpack_from(cls._ACTPRT_INFO_PACK_STR, buf, offset)
        assert cls.LACP_TLV_TYPE_ACTOR == actor_tag
        assert cls._ACTPRT_INFO_PACK_LEN == actor_length
        offset += cls._ACTPRT_INFO_PACK_LEN
        actor_state_activity = (actor_state >> 0) & 1
        actor_state_timeout = (actor_state >> 1) & 1
        actor_state_aggregation = (actor_state >> 2) & 1
        actor_state_synchronization = (actor_state >> 3) & 1
        actor_state_collecting = (actor_state >> 4) & 1
        actor_state_distributing = (actor_state >> 5) & 1
        actor_state_defaulted = (actor_state >> 6) & 1
        actor_state_expired = (actor_state >> 7) & 1
        # ------------------------------
        # Partner Information
        # ------------------------------
        (partner_tag, partner_length, partner_system_priority,
         partner_system, partner_key, partner_port_priority,
         partner_port, partner_state
         ) = struct.unpack_from(cls._ACTPRT_INFO_PACK_STR, buf, offset)
        assert cls.LACP_TLV_TYPE_PARTNER == partner_tag
        assert cls._ACTPRT_INFO_PACK_LEN == partner_length
        offset += cls._ACTPRT_INFO_PACK_LEN
        partner_state_activity = (partner_state >> 0) & 1
        partner_state_timeout = (partner_state >> 1) & 1
        partner_state_aggregation = (partner_state >> 2) & 1
        partner_state_synchronization = (partner_state >> 3) & 1
        partner_state_collecting = (partner_state >> 4) & 1
        partner_state_distributing = (partner_state >> 5) & 1
        partner_state_defaulted = (partner_state >> 6) & 1
        partner_state_expired = (partner_state >> 7) & 1
        # ------------------------------
        # Collector Information
        # ------------------------------
        (collector_tag, collector_length, collector_max_delay
         ) = struct.unpack_from(cls._COL_INFO_PACK_STR, buf, offset)
        assert cls.LACP_TLV_TYPE_COLLECTOR == collector_tag
        assert cls._COL_INFO_PACK_LEN == collector_length
        offset += cls._COL_INFO_PACK_LEN
        # ------------------------------
        # Terminator Information
        # ------------------------------
        (terminator_tag, terminator_length
         ) = struct.unpack_from(cls._TRM_PACK_STR, buf, offset)
        assert cls.LACP_TLV_TYPE_TERMINATOR == terminator_tag
        assert 0 == terminator_length
        return cls(version,
                   actor_system_priority,
                   addrconv.mac.bin_to_text(actor_system),
                   actor_key, actor_port_priority,
                   actor_port, actor_state_activity,
                   actor_state_timeout, actor_state_aggregation,
                   actor_state_synchronization, actor_state_collecting,
                   actor_state_distributing, actor_state_defaulted,
                   actor_state_expired, partner_system_priority,
                   addrconv.mac.bin_to_text(partner_system),
                   partner_key, partner_port_priority,
                   partner_port, partner_state_activity,
                   partner_state_timeout, partner_state_aggregation,
                   partner_state_synchronization,
                   partner_state_collecting, partner_state_distributing,
                   partner_state_defaulted, partner_state_expired,
                   collector_max_delay), None, buf[lacp._ALL_PACK_LEN:]

    def serialize(self, payload, prev):
        header = struct.pack(self._HLEN_PACK_STR, self._subtype,
                             self.version)
        actor = struct.pack(self._ACTPRT_INFO_PACK_STR,
                            self._actor_tag, self._actor_length,
                            self.actor_system_priority,
                            addrconv.mac.text_to_bin(self.actor_system),
                            self.actor_key,
                            self.actor_port_priority, self.actor_port,
                            self._actor_state)
        partner = struct.pack(self._ACTPRT_INFO_PACK_STR,
                              self._partner_tag, self._partner_length,
                              self.partner_system_priority,
                              addrconv.mac.text_to_bin(self.partner_system),
                              self.partner_key,
                              self.partner_port_priority,
                              self.partner_port, self._partner_state)
        collector = struct.pack(self._COL_INFO_PACK_STR,
                                self._collector_tag,
                                self._collector_length,
                                self.collector_max_delay)
        terminator = struct.pack(self._TRM_PACK_STR,
                                 self._terminator_tag,
                                 self._terminator_length)
        return header + actor + partner + collector + terminator
